#pragma once

#include <stdint.h>
#include <stdbool.h>

#if !defined VERASCO

volatile int32_t input_compile_time;
volatile int32_t input_private;
volatile int32_t input_public;
volatile int32_t output_run_time;

uint32_t
read_compile_time(void)
{
  return input_compile_time;
}

int32_t
read_public_int(int32_t min, int32_t max)
{
  __builtin_annot("input", min, max);
  return input_public;
}

uint32_t
read_private_bit(void)
{
  return input_private;
}

uint32_t
read_private_int(uint32_t nb_bits)
{
  uint32_t res = 0;
  for ( uint32_t i = 0 ; i < nb_bits ; ++i )
  {
    res = res * 2 + read_private_bit();
  }
  return res;
}

void
write_int(uint32_t v)
{
  output_run_time = v;
}

void verasco_unroll( int n ) {}

#else /* ! defined VERASCO */

#include <verasco.h>

uint32_t
read_compile_time(void)
{
  return any_int64();
}

int32_t
read_public_int(int32_t min, int32_t max)
{
  int32_t res = any_int64();
  verasco_assume ( min <= res && res <= max );
  return res;
}

uint32_t
read_private_int(uint32_t nb_bits)
{
  return any_int64();
}

void
write_int(uint32_t v)
{
  /* nothing */
}

#endif /* ! defined VERASCO */

void
assert(bool cond)
{
  if (cond) {} else { while (1); }
}

int32_t
read_char(void)
{
  return read_public_int(0, 0xFF);
}
