with import <nixpkgs> {};

let ocamlPackages = ocaml-ng.ocamlPackages_4_08; in
let coqPackages = coqPackages_8_9; in

let inherit (coqPackages) coq; in

let menhir = ocamlPackages.menhir.overrideAttrs (o: {
  src = fetchurl {
    url = "https://gitlab.inria.fr/fpottier/menhir/repository/20190626/archive.tar.gz";
    sha256 = "0nigjnskg89knyi2zj1w211mb1pvkrwfqpz9a0qbw80k3hm8gg0h";
  };
}); in

let callPackage = newScope { inherit coq ocamlPackages menhir; inherit (coqPackages) bignums; }; in

let compcert = callPackage ./nix/compcert.nix {
  tools = pkgsi686Linux.gcc;
}; in

let ppsimpl = callPackage ./nix/ppsimpl.nix {}; in

let coqprime = callPackage ./nix/coqprime.nix {}; in

stdenv.mkDerivation {
  name = "p8q";
  CompCert = "${compcert}/";
  buildInputs = [ coq ppsimpl coqprime compcert coqPackages.mathcomp-algebra_1_8 ]
    ++ (with ocamlPackages; [ ocaml findlib ocamlbuild menhir merlin ]);

  src = ./.;

  configurePhase = ''
    ln -s "${compcert}/lib/coq/${coq.coq-version}/user-contrib/compcert/" cc;
    coq_makefile -o Makefile -f _CoqProject
  '';

  enableParallelBuilding = true;

  postBuild = "make rtlc.native QapRead.native";

  installPhase = ''
    make COQLIB=$out/lib/coq/${coq.coq-version}/ install
    mkdir -p $out/bin
    mv _build/frontend/rtlc.native $out/bin/rtlc
    mv _build/frontend/QapRead.native $out/bin/qapRead
    mkdir -p $out/share
    cp ${compcert}/share/compcert.ini $out/share/
  '';
}
