(*
Warning: IMPORTANT WARNING NOTICE: CompCert 3.6 DOES NOT BELONG TO MICROSOFT - IT IS THIRD PARTY TECHNOLOGY. IT IS CLEARED ONLY FOR RESEARCH USE BY A SPECIFIC MSR RESEARCH TEAM. DO NOT USE IT UNTIL YOU HAVE CLEARED ITS USE FOR YOUR PROJECT WITH YOUR LEGAL CONTACT
*)

(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** The whole compiler and its proof of semantic preservation *)

From compcert
Require Import Compiler Linking.
Require Import String.
Import Coqlib Errors.

Local Open Scope string_scope.

Definition transf_rtl_program (tc cse cp dc: unit -> bool) (f: RTL.program) : res RTL.program :=
   OK f
   @@ print (print_RTL 0)
   @@ total_if tc (time "Tail calls" Tailcall.transf_program)
   @@ print (print_RTL 1)
  @@@ time "Inlining" Inlining.transf_program
   @@ print (print_RTL 2)
   @@ time "Renumbering" Renumber.transf_program
   @@ print (print_RTL 3)
   @@ total_if cp (time "Constant propagation" Constprop.transf_program)
   @@ print (print_RTL 4)
   @@ total_if cp (time "Renumbering" Renumber.transf_program)
   @@ print (print_RTL 5)
  @@@ partial_if cse (time "CSE" CSE.transf_program)
   @@ print (print_RTL 6)
  @@@ partial_if dc (time "Redundancy elimination" Deadcode.transf_program)
   @@ print (print_RTL 7).

Definition transf_cminor_program (tc cse cp dc: unit -> bool) (p: Cminor.program) : res RTL.program :=
   OK p
   @@ print print_Cminor
  @@@ time "Instruction selection" Selection.sel_program
  @@@ time "RTL generation" RTLgen.transl_program
  @@@ (transf_rtl_program tc cse cp dc).

Definition transf_clight_program (tc cse cp dc: unit -> bool) (p: Clight.program) : res RTL.program :=
  OK p 
   @@ print print_Clight
  @@@ time "Simplification of locals" SimplLocals.transf_program
  @@@ time "C#minor generation" Cshmgen.transl_program
  @@@ time "Cminor generation" Cminorgen.transl_program
  @@@ (transf_cminor_program tc cse cp dc).

Definition transf_c_program (tc cse cp dc: unit -> bool) (p: Csyntax.program) : res RTL.program :=
  OK p 
  @@@ time "Clight generation" SimplExpr.transl_program
  @@@ (transf_clight_program tc cse cp dc).

(** This is the list of compilation passes of CompCert in relational style.
  Each pass is characterized by a [match_prog] relation between its
  input code and its output code.  The [mkpass] and [:::] combinators,
  defined in module [Linking], ensure that the passes are composable
  (the output language of a pass is the input language of the next pass)
  and that they commute with linking (property [TransfLink], inferred
  by the type class mechanism of Coq). *)

Local Open Scope linking_scope.

Definition CompCert's_passes tc cse cp dc :=
      mkpass SimplExprproof.match_prog
  ::: mkpass SimplLocalsproof.match_prog
  ::: mkpass Cshmgenproof.match_prog
  ::: mkpass Cminorgenproof.match_prog
  ::: mkpass Selectionproof.match_prog
  ::: mkpass RTLgenproof.match_prog
  ::: mkpass (match_if tc Tailcallproof.match_prog)
  ::: mkpass Inliningproof.match_prog
  ::: mkpass Renumberproof.match_prog
  ::: mkpass (match_if cp Constpropproof.match_prog)
  ::: mkpass (match_if cp Renumberproof.match_prog)
  ::: mkpass (match_if cse CSEproof.match_prog)
  ::: mkpass (match_if dc Deadcodeproof.match_prog)
  ::: pass_nil _.

(** Composing the [match_prog] relations above, we obtain the relation
  between CompCert C sources and Asm code that characterize CompCert's
  compilation. *)

Definition match_prog tc cse cp dc : Csyntax.program -> RTL.program -> Prop :=
  pass_match (compose_passes (CompCert's_passes tc cse cp dc )).

(** The [transf_c_program] function, when successful, produces
  assembly code that is in the [match_prog] relation with the source C program. *)

Theorem transf_c_program_match:
  forall tc cse cp dc p tp,
  transf_c_program tc cse cp dc p = OK tp ->
  match_prog tc cse cp dc p tp.
Proof.
  intros tc cse cp dc p tp T.
  unfold transf_c_program, time in T. simpl in T.
  destruct (SimplExpr.transl_program p) as [p1|e] eqn:P1; simpl in T; try discriminate.
  unfold transf_clight_program, time in T. rewrite ! compose_print_identity in T. simpl in T.
  destruct (SimplLocals.transf_program p1) as [p2|e] eqn:P2; simpl in T; try discriminate.
  destruct (Cshmgen.transl_program p2) as [p3|e] eqn:P3; simpl in T; try discriminate.
  destruct (Cminorgen.transl_program p3) as [p4|e] eqn:P4; simpl in T; try discriminate.
  unfold transf_cminor_program, time in T. rewrite ! compose_print_identity in T. simpl in T.
  destruct (Selection.sel_program p4) as [p5|e] eqn:P5; simpl in T; try discriminate.
  destruct (RTLgen.transl_program p5) as [p6|e] eqn:P6; simpl in T; try discriminate.
  unfold transf_rtl_program, time in T. rewrite ! compose_print_identity in T. simpl in T.
  set (p7 := total_if tc Tailcall.transf_program p6) in *.
  destruct (Inlining.transf_program p7) as [p8|e] eqn:P8; simpl in T; try discriminate.
  set (p9 := Renumber.transf_program p8) in *.
  set (p10 := total_if cp Constprop.transf_program p9) in *.
  set (p11 := total_if cp Renumber.transf_program p10) in *.
  destruct (partial_if cse CSE.transf_program p11) as [p12|e] eqn:P12; simpl in T; try discriminate.
  destruct (partial_if dc Deadcode.transf_program p12) as [p13|e] eqn:P13; simpl in T; try discriminate.
  unfold match_prog; simpl.
  exists p1; split. apply SimplExprproof.transf_program_match; auto.
  exists p2; split. apply SimplLocalsproof.match_transf_program; auto.
  exists p3; split. apply Cshmgenproof.transf_program_match; auto.
  exists p4; split. apply Cminorgenproof.transf_program_match; auto.
  exists p5; split. apply Selectionproof.transf_program_match; auto.
  exists p6; split. apply RTLgenproof.transf_program_match; auto.
  exists p7; split. apply total_if_match. apply Tailcallproof.transf_program_match.
  exists p8; split. apply Inliningproof.transf_program_match; auto.
  exists p9; split. apply Renumberproof.transf_program_match; auto.
  exists p10; split. apply total_if_match. apply Constpropproof.transf_program_match.
  exists p11; split. apply total_if_match. apply Renumberproof.transf_program_match.
  exists p12; split. eapply partial_if_match; eauto. apply CSEproof.transf_program_match.
  exists p13; split. eapply partial_if_match; eauto. apply Deadcodeproof.transf_program_match.
  congruence.
Qed.

(** * Semantic preservation *)

Require Import MoreRTL.
Import Coqlib Smallstep.

From compcert
Require SimplExprproof
SimplLocalsproof
Cshmgenproof
Cminorgenproof
Selectionproof
RTLgenproof
Tailcallproof
Inliningproof
Renumberproof
Constpropproof
CSEproof
Deadcodeproof.

(** We now prove that the whole CompCert compiler (as characterized by the
  [match_prog] relation) preserves semantics by constructing
  the following simulations:
- Forward simulations from [Cstrategy] to [Asm]
  (composition of the forward simulations for each pass).
- Backward simulations for the same languages
  (derived from the forward simulation, using receptiveness of the source
  language and determinacy of [Asm]).
- Backward simulation from [Csem] to [Asm]
  (composition of two backward simulations).
*)

Theorem cstrategy_semantic_preservation:
  forall tc cse cp dc p tp,
  match_prog tc cse cp dc p tp ->
  forward_simulation (Cstrategy.semantics p) (RTL.semantics tp)
  /\ backward_simulation (atomic (Cstrategy.semantics p)) (RTL.semantics tp).
Proof.
  intros tc cse cp dc.
  intros p tp M. unfold match_prog, pass_match in M; simpl in M.
Ltac DestructM :=
  match goal with
    [ H: exists p, _ /\ _ |- _ ] =>
      let p := fresh "p" in let M := fresh "M" in let MM := fresh "MM" in
      destruct H as (p & M & MM); clear H
  end.
  repeat DestructM. subst tp.
  assert (F: forward_simulation (Cstrategy.semantics p) (RTL.semantics p13)).
  {
  eapply compose_forward_simulations.
    eapply SimplExprproof.transl_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply SimplLocalsproof.transf_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply Cshmgenproof.transl_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply Cminorgenproof.transl_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply Selectionproof.transf_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply RTLgenproof.transf_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply match_if_simulation. eassumption. exact Tailcallproof.transf_program_correct.
  eapply compose_forward_simulations.
    eapply Inliningproof.transf_program_correct; eassumption.
  eapply compose_forward_simulations. eapply Renumberproof.transf_program_correct; eassumption.
  eapply compose_forward_simulations.
    eapply match_if_simulation. eassumption. exact Constpropproof.transf_program_correct.
  eapply compose_forward_simulations.
    eapply match_if_simulation. eassumption. exact Renumberproof.transf_program_correct.
  eapply compose_forward_simulations.
    eapply match_if_simulation. eassumption. exact CSEproof.transf_program_correct.
  eapply match_if_simulation. eassumption. exact Deadcodeproof.transf_program_correct; eassumption.
  }
  split. auto.
  apply forward_to_backward_simulation.
  apply factor_forward_simulation. auto. eapply sd_traces. eapply MoreRTL.semantics_determinate.
  apply atomic_receptive. apply Cstrategy.semantics_strongly_receptive.
  apply MoreRTL.semantics_determinate.
Qed.

Theorem c_semantic_preservation:
  forall tc cse cp dc p tp,
  match_prog tc cse cp dc p tp ->
  backward_simulation (Csem.semantics p) (RTL.semantics tp).
Proof.
  intros tc cse cp dc.
  intros.
  apply compose_backward_simulation with (atomic (Cstrategy.semantics p)).
  eapply sd_traces; eapply MoreRTL.semantics_determinate.
  apply factor_backward_simulation.
  apply Cstrategy.strategy_simulation.
  apply Csem.semantics_single_events.
  eapply ssr_well_behaved; eapply Cstrategy.semantics_strongly_receptive.
  exact (proj2 (cstrategy_semantic_preservation _ _ _ _ _ _ H)).
Qed.

(** * Correctness of the CompCert compiler *)

(** Combining the results above, we obtain semantic preservation for two
  usage scenarios of CompCert: compilation of a single monolithic program,
  and separate compilation of multiple source files followed by linking.

  In the monolithic case, we have a whole C program [p] that is
  compiled in one run of CompCert to a whole Asm program [tp].
  Then, [tp] preserves the semantics of [p], in the sense that there
  exists a backward simulation of the dynamic semantics of [p]
  by the dynamic semantics of [tp]. *)

Theorem transf_c_program_correct:
  forall tc cse cp dc p tp,
  transf_c_program tc cse cp dc p = OK tp ->
  backward_simulation (Csem.semantics p) (RTL.semantics tp).
Proof.
  intros. apply (c_semantic_preservation tc cse cp dc). apply transf_c_program_match; auto.
Qed.

(** Here is the separate compilation case.  Consider a nonempty list [c_units]
  of C source files (compilation units), [C1 ,,, Cn].  Assume that every
  C compilation unit [Ci] is successfully compiled by CompCert, obtaining
  an Asm compilation unit [Ai].  Let [asm_unit] be the nonempty list
  [A1 ... An].  Further assume that the C units [C1 ... Cn] can be linked
  together to produce a whole C program [c_program].  Then, the generated
  Asm units can be linked together, producing a whole Asm program
  [asm_program].  Moreover, [asm_program] preserves the semantics of
  [c_program], in the sense that there exists a backward simulation of
  the dynamic semantics of [asm_program] by the dynamic semantics of [c_program].
*)

Theorem separate_transf_c_program_correct:
  forall tc cse cp dc c_units asm_units c_program,
  nlist_forall2 (fun cu tcu => transf_c_program tc cse cp dc cu = OK tcu) c_units asm_units ->
  link_list c_units = Some c_program ->
  exists asm_program,
      link_list asm_units = Some asm_program
   /\ backward_simulation (Csem.semantics c_program) (RTL.semantics asm_program).
Proof.
  intros tc cse cp dc.
  intros.
  assert (nlist_forall2 (match_prog tc cse cp dc) c_units asm_units).
  { eapply nlist_forall2_imply. eauto. simpl; intros. apply transf_c_program_match; auto. }
  assert (exists asm_program, link_list asm_units = Some asm_program /\ match_prog tc cse cp dc c_program asm_program).
  { eapply link_list_compose_passes; eauto. }
  destruct H2 as (asm_program & P & Q).
  exists asm_program; split; auto. apply (c_semantic_preservation tc cse cp dc); auto.
Qed.

Import Globalenvs.

Lemma partial_if_symbols_preserved {F V} f T (p tp: AST.program F V) :
  partial_if f T p = OK tp ->
  (forall tp, T p = OK tp -> forall s, Genv.find_symbol (Genv.globalenv tp) s = Genv.find_symbol (Genv.globalenv p) s) ->
  forall s, Genv.find_symbol (Genv.globalenv p) s = Genv.find_symbol (Genv.globalenv tp) s.
Proof.
  unfold partial_if; destruct (f _). firstorder; fail.
  intros H; inv H. auto.
Qed.

Lemma total_if_symbols_preserved {F V} f T (p: AST.program F V) :
  (forall s, Genv.find_symbol (Genv.globalenv (T p)) s = Genv.find_symbol (Genv.globalenv p) s) ->
  forall s, Genv.find_symbol (Genv.globalenv p) s = Genv.find_symbol (Genv.globalenv (total_if f T p)) s.
Proof. unfold total_if; destruct (f _); auto. Qed.

Local Opaque Csem.genv_genv.

Lemma symbols_preserved tc cse cp dc prog rtl s :
  transf_c_program tc cse cp dc prog = OK rtl ->
  Genv.find_symbol (Csem.genv_genv (Csem.globalenv prog)) s = Genv.find_symbol (Genv.globalenv rtl) s.
Proof.
  unfold transf_c_program, transf_clight_program, transf_cminor_program, transf_rtl_program; simpl.
  unfold time; simpl.
  repeat match goal with
  | |- context [ print _ _ ] => rewrite print_identity
  | |- context [ _ @@ print _ ] => rewrite compose_print_identity
  | |- context [ ?p @@@ _ = OK _ ] =>
    let X := fresh in
    destruct p eqn: X; [ simpl | discriminate ];
    revert X
  | |- context [ ?p @@ _ = OK _ ] =>
    let X := fresh in
    destruct p eqn: X; [ simpl | discriminate ];
    revert X
  end.
  intros; repeat match goal with X : OK _ = OK _ |- _ => inv X end.
  erewrite <- SimplExprproof.symbols_preserved.
  2: apply SimplExprproof.transf_program_match; eassumption.
  erewrite <- SimplLocalsproof.symbols_preserved.
  2: apply SimplLocalsproof.match_transf_program; eassumption.
  erewrite <- Cshmgenproof.symbols_preserved.
  2: apply Cshmgenproof.transf_program_match; eassumption.
  erewrite <- Cminorgenproof.symbols_preserved.
  2: apply Cminorgenproof.transf_program_match; eassumption.
  erewrite <- Selectionproof.symbols_preserved.
  2: apply Selectionproof.transf_program_match; eassumption.
  erewrite <- RTLgenproof.symbols_preserved.
  2: apply RTLgenproof.transf_program_match; eassumption.
  etransitivity.
    etransitivity.
    2: symmetry; apply Inliningproof.symbols_preserved, Inliningproof.transf_program_match; eassumption.
    apply total_if_symbols_preserved.
    apply Tailcallproof.symbols_preserved, Tailcallproof.transf_program_match.
  etransitivity.
    symmetry; apply Renumberproof.symbols_preserved, Renumberproof.transf_program_match.
  etransitivity.
    apply (total_if_symbols_preserved cp).
    apply Constpropproof.symbols_preserved, Constpropproof.transf_program_match.
  etransitivity.
  2: eapply partial_if_symbols_preserved. 2: eassumption.
  2: intros ? X; apply Deadcodeproof.symbols_preserved, Deadcodeproof.transf_program_match, X.
  etransitivity.
    apply (total_if_symbols_preserved cp).
    apply Renumberproof.symbols_preserved, Renumberproof.transf_program_match.
  eapply partial_if_symbols_preserved. eassumption.
  intros ? X; apply CSEproof.symbols_preserved, CSEproof.transf_program_match, X.
Qed.
