Require Import AbStateImpl.
Import Relation_Definitions RelationClasses.
Import String ZArith.
Import Globalenvs Integers Maps Registers Memory.
Import MoreLib Monads AbInt Expr AbState.

Section SOUNDNESS.
Let E := Errors.errmsg.

Context
  (var: Type)
  (S: Type) (sincr : relation S) `{sincr_refl: Reflexive _ sincr} `{sincr_trans: Transitive _ sincr}
  (Oracle: Type) (o: oracle var Oracle)
  (arith : Arith.t E var sincr).

Ltac crw := repeat match goal with H : _ = _ |- _ => rewrite H; simpl; clear H end.
Ltac and := repeat refine (conj _ _).

Let val := val var.
Let ab_state := (sa_state var * Oracle * S)%type.
Let incr : relation ab_state := @incr var S sincr Oracle.

Context (solution: S → (var → Z) → Prop) (Inv : S → Prop).

Hypothesis sincr_solution : ∀ σ σ', sincr σ σ' → solution σ' ⊆ solution σ.
Hint Resolve sincr_solution.
Variable ArithSound : Arith.spec arith solution Inv.

Variable prog : RTL.program.
Let ge : RTL.genv := Genv.globalenv prog.
Let sa := @sa ge var S sincr sincr_refl sincr_trans arith Oracle o.

Definition concrete_value ρ sp (x: val) : Values.val :=
  match x with
  | Const c => eval_constant ge sp c
  | Var v => ρ v
  end.

Definition sol (σ: ab_state) (ρ: val → Values.val) : Prop :=
  (∀ c, ρ (Const c) = eval_constant ge (fst (fst σ)).(sa_sp) c) ∧
  ∃ ρ', solution (snd σ) ρ' ∧ ∀ x, Values.Vint (Int.repr (ρ' x)) = ρ (Var x).

Lemma incr_sp : ∀ {σ σ' : ab_state}, incr σ σ' → sa_sp (fst (fst σ)) = sa_sp (fst (fst σ')).
Proof. intros ((a, b), c) ((a', b'), c') [ [ A B ] C ]. exact A. Qed.

Lemma incr_sol : ∀ σ σ' : ab_state, incr σ σ' → sol σ' ⊆ sol σ.
Proof.
  intros σ σ' H ρ ( Hcst & ρ' & Hsol & Hρ').
  and. intros c. rewrite (incr_sp H). apply Hcst.
  exists ρ'. destruct H; eauto.
Qed.
Hint Resolve incr_sol.

Definition ms : match_istate ab_state val :=
  λ ρ σ sp rs m,
  let '(is, _, s) := σ in
  (∀ x v, is.(sa_reg) ! x = Some v → concrete_value (ρ ∘ Var) sp v = rs # x)
∧ (∀ ofs v, get_stk is.(sa_stk) ofs = Result v → ∃ c, Mem.loadv AST.Mint32 m (Values.Val.offset_ptr sp ofs) = Some c ∧ concrete_value (ρ ∘ Var) sp v = c)
∧ (∀ s ofs v, is.(sa_global) s ofs = Result v →
              ∃ b, Genv.find_symbol ge s = Some b ∧
                   (∀ i, sp ≠ Values.Vptr b i) ∧
                   ∃ c, Mem.load AST.Mint32 m b (Ptrofs.unsigned ofs) = Some c ∧ concrete_value (ρ ∘ Var) sp v = c)
∧ (sp = is.(sa_sp))
∧ Inv s
.

Lemma ms_ext ρ σ sp rs m m':
  ms ρ σ sp rs m →
  (∀ addr, Mem.loadv AST.Mint32 m' addr = Mem.loadv AST.Mint32 m addr) →
  ms ρ σ sp rs m'.
Proof.
unfold ab_state in σ. hsplit.
intros (hreg & hstk & hglob & hsp & I) EXT. repeat refine (conj _ _); try assumption; subst.
- { intros ofs v H. destruct (hstk ofs v H) as (c & hc & hc'). exists c. split; auto. now erewrite EXT. }
- { intros s ofs v H. destruct (hglob s ofs v H) as (b & ? & ? & c & ? & ?). exists b. repeat refine (conj _ _); auto.
    exists c. split; auto. generalize (EXT (Values.Vptr b ofs)). simpl. congruence. }
Qed.

Lemma add0 x :
  Int.add (Int.add x (Int.repr Int.half_modulus)) (Int.repr Int.half_modulus) = x.
Proof.
  rewrite Int.add_assoc.
  rewrite ! Int.add_unsigned.
  do 2 change (Int.unsigned (Int.repr Int.half_modulus)) with Int.half_modulus.
  rewrite Z.add_0_r. apply Int.repr_unsigned.
Qed.

Lemma eval_sound rs m sp :
  ∀ e σ v σ' INCR,
  eval arith o e σ = STm.Result v σ' INCR →
  ∀ ρ,
    sol σ' ρ →
    ms ρ σ sp rs m →
    rtl_eval_expr ge sp rs m e tt = STm.Result (concrete_value (ρ ∘ Var) sp v) tt I ∧
    ms ρ σ' sp rs m ∧
    fst (fst σ') = fst (fst σ).
Proof.
  induction e; intros ((σ & or) & s) res ((σ' & or') & s') INCR EVAL ρ SOL (MSr & MSs & MSg & -> & INV); simpl in *; hsplit; simpl in *; repeat subst; hsplit; subst.
  - pose proof (MSr v) as K. flatten hsplit. subst. specialize (K _ eq_refl). rewrite K.
    repeat (refine (conj _ _)); auto.
  - repeat (refine (conj _ _)); auto.
  - destruct x as [ [?|?|?] | x ]; hsplit; simpl in *; subst;
     edestruct IHe; eauto; simpl; eauto; clear IHe; simpl in *; hsplit; subst.
    +
    generalize (MSs ofs _ H1). intros; hsplit; subst.
    repeat (refine (conj _ _)); auto.
    unfold STm.bind. now crw.
    +
    generalize (MSg _ _ _ H1). intros; hsplit; subst.
    repeat (refine (conj _ _)); auto.
    unfold STm.bind. now crw.
  - simpl in *.
    edestruct IHe1; eauto; simpl; eauto. clear IHe1. simpl in *. hsplit. subst.
    edestruct IHe2; eauto; simpl; eauto. clear IHe2. simpl in *. hsplit. subst.
    unfold binop in EVAL.
    destruct (two_consts _ _) as [ ( cst & cst' & -> & ->) | Hcst ].

    + hsplit. subst. hsplit. subst.
Opaque STm.left.
simpl in *.
unfold STm.bind. crw. destruct op; simpl in *; hsplit; subst.
and; eauto. flatten hsplit; subst; simpl; auto.
  destruct (sa_sp _); try easy; simpl. now rewrite Ptrofs.add_assoc, (Ptrofs.add_commut ofs).
  destruct (Genv.find_symbol ge _). now rewrite Ptrofs.add_commut. easy.
  destruct (sa_sp _); try easy; simpl. now rewrite Ptrofs.add_assoc, (Ptrofs.add_commut ofs).
  destruct (Genv.find_symbol ge _); try easy; simpl; now rewrite Ptrofs.add_commut.
and; eauto. flatten hsplit; subst; simpl; auto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
destruct cst; hsplit. simpl in *. subst. and; eauto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
f_equal. simpl; flatten hsplit; subst; simpl; auto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
f_equal. destruct signed; simpl; flatten hsplit; subst; simpl; auto.
destruct cst; hsplit. destruct cst'; hsplit. subst. simpl in *. and; eauto.
destruct signed.
  unfold Values.Val.cmp. simpl. destruct (Int.cmp _ _ _); auto.
  unfold Values.Val.cmpu. simpl. destruct (Int.cmpu _ _ _); auto.

  +
    set (op' := op); unfold op' in *.
    (destruct (is_slt0 _ _) as [ (v & ?) | _ ]; [| destruct op; flatten hsplit]); hsplit; simpl in *; subst;
    hsplit; repeat subst;
    unfold STm.bind; simpl; crw;
repeat match goal with
  | H : o_hint _ _ _ = STm.Result _ _ _ |- _ => clear H
  | SOL : sol (_, ?σ) ?ρ |- _ =>
    let ρ' := fresh "ρ'" in
    let SOL' := fresh "SOL" in
    destruct SOL as ( ? & ρ' & SOL' & ?);
    simpl in *
  | A : Arith.of_int _ _ _ = STm.Result _ ?σ _, K: solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto 6;
    (destruct (Arith.of_int_sound ArithSound A SOL); [ now eauto | ]);
    clear A
  | A : Arith.add _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ' ?ρ |- _ =>
    let SOL' := fresh "SOL" in
    assert (solution σ ρ) as SOL' by eauto;
    (destruct (Arith.add_sound ArithSound A SOL'); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.sub _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.sub_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.mul _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.mul_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
             (*
  | A : Arith.lnot _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.lnot_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
*)
  | A : Arith.land _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.land_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.lor _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.lor_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.lxor _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.lxor_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.shru _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.shru_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.cmp _ _ _ _ _ _ = STm.Result _ ?σ _, SOL : solution ?σ ?ρ |- _ =>
    (destruct (Arith.cmp_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.is_neg _ _ _ = STm.Result _ ?σ _, K : solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto;
    (destruct (Arith.is_neg_sound ArithSound sincr_solution A SOL); [ now eauto | ]);
    clear A;
    simpl in *
end; subst;

and; eauto; crw; simpl;
try unfold Values.Val.cmp, Values.Val.cmpu;
simpl;
repeat match goal with
| H : ∀ x, Values.Vint (Int.repr (?ρ x)) = ?ρ' (Var x) |- _ => rewrite <- ! H; clear H
| H : _ = Int.repr (?ρ ?x) |- _ => rewrite <- H in * ; clear x H
| H : Int.repr (?ρ ?x) = _ |- _ => rewrite H in * ; clear x H
end;
simpl;
crw;
try (exact eq_refl).

rewrite add0.
now match goal with |- context [ if ?b then _ else _ ] => destruct b end.
f_equal. f_equal. apply Int.shl_mul.
now repeat match goal with |- context [ if ?b then _ else _ ] => destruct b; simpl end.
now repeat match goal with |- context [ if ?b then _ else _ ] => destruct b; simpl end.
now repeat match goal with |- context [ if ?b then _ else _ ] => destruct b; simpl end.
Qed.
Arguments eval_sound rs m sp [e σ v σ' INCR] _ [ρ] _ _.

Lemma key_of_offset_inj ofs ofs' :
  key_of_offset ofs = key_of_offset ofs' →
  ofs = ofs'.
Proof.
  clear.
  unfold key_of_offset. generalize (proj1 (Ptrofs.unsigned_range ofs)), (proj1 (Ptrofs.unsigned_range ofs')).
  intros H H' EQ. apply unsigned_inj.
  destruct (Ptrofs.unsigned ofs), (Ptrofs.unsigned ofs'); Psatz.lia.
Qed.

Local Ltac t := repeat (simpl in *; hsplit; subst).

Theorem sa_sound : state_abstraction_spec sa prog sol ms.
Proof.
  constructor.
- intros dst.
  intros e s s' Hincr ASSN. simpl in ASSN. hsplit. subst.
  intros ρ SOL rs m sp MS'. simpl in *. repeat subst. simpl in *.
  destruct (eval_sound _ _ _ ASSN0 SOL MS') as (EVAL & (Reg & Stk & Global) & ?); hsplit; repeat subst.
  eexists; split. eassumption. and; auto.
  intros x1 v. rewrite PTree.gsspec, Regmap.gsspec.
  destruct (Coqlib.peq _ _). subst. congruence.
  intuition.
- (* forget *)
  simpl. intros dst σ σ' Hincr H ρ (Hc & ρ' & Hsol' & Hρ') rs m sp MS. hsplit. simpl in *. hsplit.
  repeat refine (conj _ _); eauto.
  intros x v. rewrite PTree.grspec. destruct PTree.elt_eq as [EQ|NE]; intros GET. hsplit. rewrite Regmap.gso; eauto.
- (* store *)
  intros dst e σ σ' Hincr H ρ Hsol rs m sp MS v a m' Hv Ha Hm'. hsplit. simpl in *. hsplit. subst. flatten hsplit. simpl in *.
  repeat match goal with
  | EV: eval _ _ _ _ = STm.Result _ ?σ' _, ρ : AbStateImpl.val _ → _ |- _ =>
    let SOL := fresh "SOL" in
    assert (sol σ' ρ) as SOL by eauto;
    (edestruct (λ rs m sp, eval_sound rs m sp EV SOL); [ now simpl; eauto |]; clear EV);
    hsplit; subst
    end.
  simpl in *. subst.
  unfold ge in *.
  repeat match goal with H : rtl_eval_expr _ _ _ _ ?e _ = _, K : rtl_eval_expr _ _ _ _ ?e _ = _ |- _ => rewrite K in H end; hsplit; repeat subst.
  destruct (sa_sp σ); simpl in *; hsplit.
  and; auto. simpl in *.
  + unfold get_stk. intros ofs0 v EQ. hnf. simpl. rewrite PTree.gsspec in EQ. destruct (Coqlib.peq _ _) as [EQ' | NE'].
    * apply key_of_offset_inj in EQ'. apply result_eq in EQ. subst.
      eexists. split. 2: reflexivity.
      erewrite Mem.load_store_same. 2: eassumption.
      destruct v; try easy. destruct c; try easy. simpl. destruct (Genv.find_symbol ge s); easy.
      simpl; f_equal.
      destruct SOL as (? & ρ' & SOL & Hρ').
      now rewrite <- Hρ'.
    * assert (ofs0 ≠ ofs) as NE. exact (λ EQ, NE' (f_equal _ EQ)). clear NE'.
      destruct (MS0 ofs0 v EQ) as (c & Hload & Hc).
      exists c. split. 2: easy.
      erewrite Mem.load_store_other; eauto. simpl. right.
      assert (Ptrofs.unsigned (Ptrofs.add i ofs0) ≠ Ptrofs.unsigned (Ptrofs.add i ofs)).
      { intros K. apply unsigned_inj, add_inj in K. easy. }
      destruct (Mem.load_valid_access _ _ _ _ _ Hload) as [_ (q & Q)]. simpl in Q. rewrite Q in *.
      destruct (Mem.store_valid_access_3 _ _ _ _ _ _ Hm') as [_ (q' & Q')]. simpl in Q'. rewrite Q' in *.
      Psatz.lia.
  + intros s ofs0 v K. destruct (MS1 s ofs0 v K) as (b' & Hb & H' & c & Hload & Hc).
    exists b'. split. easy. split. easy. exists c. split. 2: easy.
    erewrite Mem.load_store_other; eauto. left.
    clear -H'. intros ->; eapply H'; reflexivity.
- intros dst κ n ofs σ v σ' Hincr H ? ρ (? & ρ' & Hsol & Hρ') rs m sp MS. simpl in *. t. simpl in *. t. and.
  + intros x0 v0. rewrite PTree.gsspec, Regmap.gsspec.
    destruct (Coqlib.peq _ _). subst. intros H'. hsplit. subst.
    now simpl;
    match goal with |- context[ ρ (Var ?u) ] => rewrite <- (Hρ' u) end.
    flatten t; try easy. auto.
  + intros; flatten t; eauto.
  + intros; flatten t; eauto.
  + easy.
  + now match goal with H : Arith.input _ _ _ _ = _ |- _ => refine (Arith.input_sound ArithSound H _) end.
  + now match goal with |- context[ ρ (Var ?u) ] => destruct (Hρ' u) end.
- simpl. intros κ n ofs src σ v σ' Hincr H H0 ρ (Hc & ρ' & Hsol' & Hρ') rs m sp MS. hsplit. repeat subst.
  {
    simpl in *.
    destruct ((sa_reg _) ! src) eqn: K; [ | hsplit]. rewrite <- (proj1 MS _ _ K). t.
    flatten t; simpl in *.
    + rewrite Hc. and; auto.
    + repeat rewrite <- (Hρ').
      destruct (Arith.output_sound ArithSound H Hsol'). easy.
      and; auto.
      congruence.
  }
- simpl. intros low hi σ σ' Hincr H ρ (Hc & ρ' & Hsol' & Hρ') rs m sp MS. hsplit. simpl in *. hsplit.
  and; auto.
- simpl. intros r σ v σ' H H0 ρ (Hc & ρ' & Hsol' & Hρ') sp rs m MS. hsplit. simpl in *.
  flatten_eq hsplit; subst. rewrite Hc. match goal with H : _ |- _ => exact (conj (MS _ _ H) I) end.
- intros m₀ σ₁ H ρ Hsol m₁ stk b fd H1 H2 H3 H4. simpl in *. hsplit. subst. simpl in *. and; simpl.
  + intros x v H. rewrite PTree.gempty in H. easy.
  + intros ofs v H. unfold get_stk in H. rewrite PTree.gempty in H. hsplit.
  + intros s ofs v H. flatten_eq hsplit. subst. eexists. and. eassumption.
    intros i0 K. inversion K; clear K; subst.
    eapply Genv.find_symbol_not_fresh in Heqo0; eauto.
    apply Mem.alloc_result in H2. subst. hnf in Heqo0. rewrite BinPos.Pos.compare_refl in Heqo0. discriminate.
    eexists. and. eapply Mem.load_alloc_other; eassumption. easy.
  + f_equal. eapply Mem.alloc_result; eassumption.
  + apply (Arith.init_sound ArithSound).
- intros e σ v σ' H H0 ρ Hsol rs m sp. hsplit. simpl in *. hsplit. simpl in *. subst.
  flatten_eq t;
  intros MS;
  repeat (match goal with
  | H : eval _ _ _ _ = STm.Result _ ?σ' _ |- _ =>
    let SOL := fresh in
    assert ( sol σ' ρ ) as SOL by eauto;
    pose proof (eval_sound _ _ _ H SOL MS); clear H SOL
  | H : sol _ _ |- _ => unfold sol in H
  | H : Arith.to_int _ _ _ = STm.Result _ ?σ' _, SOL: solution ?σ' ?ρ |- _ =>
    pose proof (Arith.to_int_sound ArithSound H SOL); clear H
  end; t); intuition.
  rewrite Heqo0 in H0. auto. subst.
  rewrite Hsol1. eauto.

- unfold sa_assert. simpl. intros x σ σ' Hincr H ρ (Hc & ρ' & Hsol' & Hρ') rs m sp MS. hsplit. simpl in *. hsplit.
  flatten_eq hsplit; subst; simpl in *; hsplit; repeat subst;
  repeat match goal with
  | A : Arith.of_int _ _ _ = STm.Result _ ?σ _, K: solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto 6;
    (destruct (Arith.of_int_sound ArithSound A SOL); [ now eauto | ]);
    clear A
  | A : Arith.add _ _ _ _ = STm.Result _ ?σ _, K : solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto 6;
    (destruct (Arith.add_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.assert _ _ _ = STm.Result _ ?σ _, K : solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto 6;
    (destruct (Arith.assert_sound ArithSound A SOL); [ now eauto |]);
    clear A
  | A : Arith.assert_eq _ _ _ _ = STm.Result _ ?σ _, K : solution ?σ _ |- _ =>
    (destruct (Arith.assert_eq_sound ArithSound A K); [ now eauto |]);
    clear A
  | A : Arith.sub _ _ _ _ = STm.Result _ ?σ _, K : solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto;
    (destruct (Arith.sub_sound ArithSound A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  | A : Arith.is_neg _ _ _ = STm.Result _ ?σ _, K : solution ?σ' ?ρ |- _ =>
    let SOL := fresh "SOL" in
    assert ( solution σ ρ ) as SOL by eauto;
    (destruct (Arith.is_neg_sound ArithSound sincr_solution A SOL); [ now eauto | ]);
    clear A;
    simpl in *
  end;
  split; eauto;
  unfold U.ret, STm.ret; unfold STm.bind; simpl; f_equal;
  repeat match goal with | EQ : _ ! ?v = Some _ |- context [ ?rs # ?v ] => rewrite <- (MS _ _ EQ) end;
  simpl; try rewrite <- ! Hρ';
  unfold Values.Val.cmp, Values.Val.cmpu; simpl;
  repeat match goal with
  | H : Int.repr (?ρ ?x) = _ |- _ => rewrite H in *; clear H
  | H : _ = Int.repr (?ρ ?x) |- _ => rewrite <- H in *; clear H
  end.
  rewrite H; reflexivity.
  rewrite H1; reflexivity.
  rewrite H0; reflexivity.
  rewrite add0 in H8.
  destruct (Int.lt _ _). reflexivity. discriminate.
  rewrite add0 in H8.
  destruct (Int.lt _ _). reflexivity. discriminate.
  rewrite H0; reflexivity.

- intros [[σ u] x] r H ρ (_ & ρ' & Hsol & Q) rs m sp [ K _ ].
  simpl in *.
  specialize (K r). destruct ((sa_reg _) ! _) as [ v | ]. 2: easy.
  specialize (K _ eq_refl). rewrite <- K. clear K.
  destruct v as [ [?|?|?] | v ]; simpl; try easy.
  now rewrite <- Q.

Qed.

End SOUNDNESS.

Arguments sa_sound [var S sincr] {_ _} [Oracle] o {arith solution Inv} _ _ _.
