Require Import Utf8 BinPos.
Require Import RField.
Import EquivDec.

Unset Elimination Schemes.
Set Implicit Arguments.

Local Open Scope positive_scope.
Local Open Scope algebra_scope.

Module Type LinearCombinationType.
  Parameter t : Type → Type.
  Definition var : Type := positive.
  Parameter zero : ∀ f, t f.
  Parameter single : ∀ f, var → f → t f.
  Parameter add : ∀ f feq `{EqDec f feq} `{Plus f} `{Zero f}, t f → t f → t f.
  Parameter sub : ∀ f feq `{EqDec f feq} `{Plus f} `{Zero f} `{Opposite f}, t f → t f → t f.
  Parameter muln : ∀ f feq `{EqDec f feq} `{Zero f} `{Mult f}, f → t f → t f.

  Parameter fold : ∀ f feq `{EqDec f feq} `{Plus f} `{Zero f} (A : Type), (A → var → f → A) → t f → A → A.

  Parameter is_zero : ∀ f feq `{EqDec f feq} `{Zero f}, t f → bool.
  Parameter is_single : ∀ f feq `{EqDec f feq} `{Zero f}, var → t f → option f.

  (* Specification *)
  Parameter den : ∀ f `{Plus f} `{Zero f} `{Mult f}, var → (var → f) → t f → f.

  Axiom den_zero : ∀ f `(Field f) ω ρ,
    den ω ρ (zero f) === 0.
  Axiom den_single : ∀ f `(Field f) ω ρ,
    ∀ x n, x <= ω → den ω ρ (single x n) === ρ x × n.
  Axiom den_add : ∀ f `(Field f) ω ρ,
    ∀ a b: t f, den ω ρ (add a b) === den ω ρ a + den ω ρ b.
  Axiom den_sub : ∀ f `(Field f) ω ρ,
    ∀ a b: t f, den ω ρ (sub a b) === den ω ρ a − den ω ρ b.
  Axiom den_muln : ∀ f `(Field f) ω ρ,
    ∀ (n: f) (a: t f), den ω ρ (muln n a) === n × den ω ρ a.

  Axiom den_is_zero : ∀ f `(Field f) ω ρ,
      ∀ a: t f, is_zero a = true → den ω ρ a === 0.
  Axiom den_is_single : ∀ f `(Field f) ω ρ,
      ∀ (a: t f) (x: var) (n: f), x <= ω → is_single x a = Some n → den ω ρ a === ρ x × n.
End LinearCombinationType.
