{ stdenv, fetchFromGitHub, tools, coq, ocamlPackages, coq2html, menhir }:

stdenv.mkDerivation {
  pname = "compcert";
  version = "3.6";

  src = fetchFromGitHub {
    owner = "AbsInt";
    repo = "CompCert";
    rev = "v3.6";
    sha256 = "1k9xhj7fgllhf7bn7rp3w6zfvs4clglnc4w39zp4678hrwvdcpha";
   };

  buildInputs = [ coq coq2html menhir ]
  ++ (with ocamlPackages; [ ocaml findlib ]);

  enableParallelBuilding = true;

  configurePhase = ''
    substituteInPlace ./configure --replace '{toolprefix}gcc' '{toolprefix}cc'
    ./configure -prefix $out -toolprefix ${tools}/bin/ '' +
    (if stdenv.isDarwin then "ia32-macosx" else "ia32-linux");

  installTargets = "documentation install";

  postInstall = ''
    mkdir -p $out/share/doc/compcert
    mv doc/html $out/share/doc/compcert/
    rm cparser/Lexer.mll cparser/pre_parser.mly
    rm lib/Tokenize.mll lib/Readconfig.mll lib/Responsefile.mll
    find . \( -name '*.cmi' -o -name '*.cmo' -o -name '*.cmx' -o -name '*.o' \) -delete
    mkdir -p $out/lib/coq/${coq.coq-version}/user-contrib/compcert/
    mv backend cfrontend common cparser debug driver flocq x86 x86_32 lib MenhirLib \
      $out/lib/coq/${coq.coq-version}/user-contrib/compcert/
  '';

  meta = {
    homepage = "http://compcert.inria.fr/";
    license = stdenv.lib.licenses.inria-compcert;
  };

}
