Require Import AbInt QAP.
Require Import LinearCombination.
Import Bool Sumbool ZArith Morphisms EquivDec RelationClasses.
Import MoreLib Monads Interval LargePrime.
Import Maps Integers.
Import RField.

Set Implicit Arguments.

Import Interval.Notations.
Local Open Scope interval_scope.
Local Open Scope Z_scope.


Section FIELD.

Context f feq fequiv feqdec fplus fzero fopp fmult fone finv `(F: @Field f feq fequiv feqdec fplus fzero fopp fmult fone finv).
Add Field relevent : (field_of_field F).
Context (f_of_z : Z → f)
        (f_of_z_zero : f_of_z 0 = zero)
        (f_of_z_one : f_of_z 1 = one)
.

Local Open Scope algebra_scope.

Notation "ω ⟦ x ⟧ y" := (LC.den ω y x) (at level 20, no associativity).

Context (project: f → interval → option Z)
        (project_in : ∀ x r z, project x r = Some z → z ∈ r)
        (project_proper: Proper (feq ==> eq ==> eq) project)
        (project_const : ∀ x, project (f_of_z x) (const x) = Some x)
        (project_zero : project 0 Interval.bit = Some Z0)
        (project_one : project 1 Interval.bit = Some (Zpos xH))
        (project_one' : ∀ x, project x (const 1) ≠ None → x === 1)
        (project_add :
           ∀ x xr x' y yr y',
             project x xr = Some x' →
             project y yr = Some y' →
             (size (Interval.add xr yr) < Npos large_prime)%N →
             project (x + y) (Interval.add xr yr) = Some (x' + y')%Z)
        (project_sub :
           ∀ x xr x' y yr y',
             project x xr = Some x' →
             project y yr = Some y' →
             (size (Interval.sub xr yr) < Npos large_prime)%N →
             project (x − y) (Interval.sub xr yr) = Some (x' - y')%Z)
        (project_muln :
           ∀ x' y yr y',
             project y yr = Some y' →
             (size (muln x' yr) < Npos large_prime)%N →
             project (f_of_z x' × y) (Interval.muln x' yr) = Some (x' * y'))
        (project_mul :
           ∀ x xr x' y yr y',
             project x xr = Some x' →
             project y yr = Some y' →
             (size (mul xr yr) < Npos large_prime)%N →
             project (x × y) (Interval.mul xr yr) = Some (x' * y'))
        (project_fun :
           ∀ x xr xr' a a',
             project x xr = Some a →
             project x xr' = Some a' →
             Interval.leb (Interval.sub xr xr') {| imin := - Int.max_unsigned; imax := Int.max_unsigned |} →
             a = a')
        (project_inj :
           ∀ x xr y yr a b,
             project x xr = Some a →
             project y yr = Some b →
             x =/= y → a ≠ b)
        (project_subrange : ∀ (x: f) (r: interval) (m: Z) (a b: Z),
           project x r = Some a →
           project x {| imin := 0; imax := m |} = Some b →
           Interval.leb r {| imin := 0; imax := m |} →
           0 <= m < Zpos large_prime →
           a = b)
.

Lemma project_is_zero a :
  project a bit = Some 0%Z → a === 0.
Proof.
  intros H.
  pose proof (project_inj H project_zero) as Q.
  destruct (feqdec a 0) as [ E | K ]. exact E. elim (Q K eq_refl).
Qed.

Lemma project_is_one a :
  project a bit = Some 1%Z → a === 1.
Proof.
  intros H.
  pose proof (project_inj H project_one) as Q.
  destruct (feqdec a 1) as [ E | K ]. exact E. elim (Q K eq_refl).
Qed.

(* A putative solution to a qap *)
Definition valuation : Type := positive → f.

Definition solution ω (q: cqap f) (β: valuation) : Prop :=
    (
    ∀ w r,
      q.(meta_data) ! w = Some (Bank_Shared r) →
      project (β w) r ≠ None
    )
    ∧
    (
    ∀ r,
      (r < q.(next_root))%positive →
      ∀ v w y,
      q.(cV) ! r = Some v →
      q.(cW) ! r = Some w →
      q.(cY) ! r = Some y →
      ω⟦v⟧β × ω⟦w⟧β === ω⟦y⟧β
    )
    ∧
    ((q.(next_wire) <= Pos.succ ω)%positive)
.


Local Notation lc_of_value := (lc_of_value (f_of_z:=f_of_z)).
Local Notation lc_of_binary_repr := (lc_of_binary_repr (f_of_z:=f_of_z)).
Local Notation get_binary_decomposition := (get_binary_decomposition (f_of_z:=f_of_z)).
Local Notation get_lc := (get_lc (f_of_z:=f_of_z)).

Definition get_env (e: M.t (value f)) (v: positive) :=
  match e ! v with
  | Some val => Some (lc_of_value val)
  | None => None
  end.

Lemma get_env_Some e v lc:
  get_env e v = Some lc <-> (exists val, e ! v = Some val /\ lc_of_value val = lc).
Proof.
  unfold get_env. case (e ! v); split; intros; hsplit. eauto. congruence.
Qed.

Corollary get_env_Some_simpl e v val
  (H:e ! v = Some val) : get_env e v = Some (lc_of_value val).
Proof. rewrite get_env_Some. exists val; intuition. Qed.

Lemma get_env_empty i :
  get_env (M.empty _) i = None.
Proof. unfold get_env. now rewrite M.gempty. Qed.

Lemma get_env_gss i e v :
  get_env (M.set i v e) i = Some (lc_of_value v).
Proof. unfold get_env. now rewrite M.gss. Qed.

Lemma get_env_gso i j e v (H:i ≠ j) :
  get_env (M.set j v e) i = get_env e i.
Proof. unfold get_env. now rewrite M.gso. Qed.

Lemma get_env_sspec i j e v :
  get_env (M.set j v e) i = if (equiv_dec i j) then Some (lc_of_value v) else get_env e i.
Proof. destruct (equiv_dec i j) as [ -> | NE ]. apply get_env_gss. apply get_env_gso, NE. Qed.

Definition cqap_solution (q: cqap f) (ρ: positive → Z) : Prop :=
  ∃ (β : valuation) ω,
    solution ω q β
    ∧
    (
      ∀ v, ∀ val, get_env q.(env) v = Some val → ∀ z, project (ω⟦fst val⟧β) (snd val) = Some z → z = ρ v
    )
.

Lemma in_md_inv (q: cqap f) x y :
  (meta_data q) ! x = Some y →
  (x < next_wire q)%positive.
Proof.
  intros H. assert (x < next_wire q ∨ next_wire q <= x)%positive as OR by Psatz.lia. destruct OR as [K | K]. easy.
  generalize (proj1 (wf q) _ K). congruence.
Qed.
Hint Resolve in_md_inv.

Lemma in_env_inv (q: cqap f) x y :
  (env q) ! x = Some y →
  (x < next_var q)%positive.
Proof.
  intros H. assert (x < next_var q ∨ next_var q <= x)%positive as OR by Psatz.lia. destruct OR as [K | K]. easy.
  generalize (proj2 (wf q) _ K). congruence.
Qed.
Hint Resolve in_env_inv.

Lemma get_env_inv (q: cqap f) x y :
  get_env (env q) x = Some y →
  (x < next_var q)%positive.
Proof.
  unfold get_env. case_eq ((env q) ! x); eauto; discriminate.
Qed.
Hint Resolve get_env_inv.

Import PP.PpsimplZ.

Lemma solution_monotone (q q': cqap f) (hIncr: incr q q') :
  ∀ ω, solution ω q' ⊆ solution ω q.
Proof.
destruct hIncr as (hw & hr & Hv & hV & hW & hY & hmd & henv).
intros ω β (h & hv & hω). repeat refine (conj _ _).
- { intros w r H. rewrite hmd in H. apply (h _ _ H). eauto. }
- { intros r LT v w y V W Y. rewrite hV in V. rewrite hW in W. rewrite hY in Y.
    apply (hv _ (Pos.lt_le_trans _ _ _ LT hr) _ _ _ V W Y).
    exact LT. exact LT. exact LT. }
- { clear - hω hw. ppsimpl. Psatz.lia. }
Qed.
Arguments solution_monotone [q q'] _ _ _ _.

Lemma cqap_monotone_precise (q q': cqap f) (hIncr: incr q q') :
  ∀ ρ β ω, solution ω q' β →
           (∀ (v : positive) (val : Lc f * interval),
              get_env (env q') v = Some val
              → ∀ z : Z, project (ω ⟦ fst val ⟧ β) (snd val) = Some z → z = ρ v) →
           solution ω q β
                    ∧ (∀ (v : positive) (val : Lc f * interval),
                         get_env (env q) v = Some val
                         → ∀ z : Z, project (ω ⟦ fst val ⟧ β) (snd val) = Some z → z = ρ v).
Proof.
  intros ρ β ω hβ hp.
  refine (conj (solution_monotone hIncr ω _ hβ) _).
  destruct hIncr as (hw & hr & Hv & hV & hW & hY & hmd & henv).
  { intros v val H. apply (hp v val). unfold get_env. rewrite <- henv. exact H. eauto. }
Qed.

Lemma cqap_monotone (q q': cqap f) (hIncr: incr q q') :
  cqap_solution q' ⊆ cqap_solution q.
Proof.
  intros ρ (β & ω & hβ & hp).
  exists β, ω. eauto using cqap_monotone_precise.
Qed.

Definition post ω e c nv (ρ: valuation) : Prop :=
  ∀ v, (∀ val,
    get_env e v = Some val →
    (∃ p, project (ω⟦ fst val ⟧ ρ) (snd val) = Some p ∧
          (∀ bs, e ! v = Some (inr bs) →
                 (length bs ≤ Int.wordsize) ∧
                List.Forall (fun b => project (ω⟦ b ⟧ ρ) Interval.bit ≠ None) bs))) ∧
    (∀ bs, c ! v = Some bs →
           (length bs ≤ Int.wordsize) ∧
           List.Forall (fun b => project (ω⟦ b ⟧ ρ) Interval.bit ≠ None) bs ∧
           ∃ val p p', get_env e v = Some val ∧
            project (ω⟦ fst val ⟧ ρ) (snd val) = Some p ∧
            project (ω⟦ lc_of_binary_repr bs ⟧ ρ) (range_of_binary_repr bs) = Some p' ∧
             Int.repr p = Int.repr p') ∧
    ((nv <= v)%positive → c ! v = None).

Definition invariant ω q : Prop :=
  solution ω q ⊆ post ω q.(env) q.(bcache) q.(next_var).

Lemma project_range_ex x r z:
  project x r = Some z → ∃ i, i ∈ r.
Proof. exists z. eapply project_in; eassumption. Qed.
Hint Resolve project_range_ex.

Lemma lt_ne x y : (x < y)%positive → x ≠ y.
Proof. Psatz.lia. Qed.

Lemma lc_of_f_denote q ω β x :
  solution ω q β →
  project (ω⟦lc_of_f (f_of_z x)⟧ β) (Interval.const x) = Some x.
Proof.
  intros (h & hp).
  unfold lc_of_f. destruct (equiv_dec _ _) as [hz|hnz].

  rewrite (LC.den_zero F), <- hz. apply project_const.
  pose proof (project_one' (h _ _ (proj2 (q.(md_invariant))))) as h1.
  rewrite (LC.den_single F), h1, left_identity. apply project_const. exact (Pos.le_1_l _).
Qed.

Lemma lc_of_one ω q β : solution ω q β → ω ⟦lc_of_f 1⟧ β === 1.
Proof.
  intros (h & _). unfold lc_of_f.
  pose proof (project_one' (h _ _ (proj2 (q.(md_invariant))))) as h1.
  rewrite if_dec_false, (LC.den_single F), right_identity. exact h1. exact (Pos.le_1_l _).
  apply field_not_trivial.
Qed.

Lemma fresh_lc_invariant e c nv val X :
  let lc := lc_of_value val in
  (∀ ω ρ, X ω ρ → project (ω ⟦fst lc⟧ρ) (snd lc) ≠ None) →
  (∀ ω ρ bs, X ω ρ → val = inr bs → (length bs ≤ Int.wordsize)) →
  (∀ ω ρ bs, val = inr bs → X ω ρ → List.Forall (fun b => project (ω⟦ b ⟧ ρ) Interval.bit ≠ None) bs) →
  (∀ ω, X ω ⊆ post ω e c nv) →
  ∀ ω, X ω ⊆ post ω (M.set nv val e) c (Pos.succ nv).
Proof.
  unfold post in *. intros H H'' H' H1 ω β INV v'.
  destruct (H1 _ _ INV v') as (H10 & H11 & H20).
  split; [ |split].
  - intros r' K. rewrite get_env_Some in K. destruct K as [val0 [K1 K2]].
    rewrite M.gsspec in K1. destruct (Coqlib.peq v' nv); hsplit; subst; simpl.
    * rewrite K2 in H. destruct (not_none (H _ _ INV)) as [p Hp].
      exists p. split; auto.
      rewrite M.gss. intros bs Hbs. inversion Hbs. subst val0. clear Hbs.
      split.
      + apply (H'' _ _ _ INV); auto.
      + apply H'; auto.
    * rewrite M.gso; auto.
      apply (H10 (r', r'0)). unfold get_env. rewrite K1, K2. reflexivity.
  - intros bs Hbs. destruct (H11 _ Hbs) as (H100 & H12 & val0 & p & p' & H13 & H14 & H15 & H16).
    repeat split; auto.
    rewrite get_env_Some in H13. destruct H13 as [val1 [K1 K2]].
    destruct (Coqlib.peq v' nv); hsplit; subst; simpl in *.
    * rewrite H20 in Hbs; [discriminate|clear; Psatz.lia].
    * exists (lc_of_value val1), p, p'.
      unfold get_env. rewrite M.gso; auto. rewrite K1, K2. simpl.
      repeat split; auto.
  - intro Hnv. apply H20. clear -Hnv. Psatz.lia.
Qed.
Arguments fresh_lc_invariant [e c nv val X] _ _ _ _ _ _ _ _.

Lemma sub_range L U a b x x' y y' :
  Interval.leb (Interval.sub a b) {| imin := L; imax := U |} = true →
  project x a = Some x' →
  project y b = Some y' →
  L <= x' - y' <= U.
Proof.
  intros LE ha hb.
  generalize (Interval.leb_denote _ _ LE _ (Interval.sub_correct _ _ _ _ (project_in ha) (project_in hb))).
  clear.
  unfold Interval.denote, is_true. simpl. rewrite Bool.andb_true_iff, ! Z.leb_le.
  exact id.
Qed.
Hint Resolve sub_range.

Let π₃ {A B C} (x: A ∧ B ∧ C) : B := proj1 (proj2 x).

Ltac clean_field :=
  repeat match goal with
  | H : context[ ?a + - 0 ] |- _ => fold (a − 0) in H; rewrite right_identity in H
  | H : context[ ?a + - ?a ] |- _ => rewrite (commutativity a), left_inverse in H
  | H : context[ ?a − ?a ] |- _ => unfold RField.sub in H; rewrite (commutativity a), left_inverse in H
  | H : context[ ?a × 0 ] |- _ => rewrite (commutativity a), left_absorb in H
  | H : context[ 0 × _ ] |- _ => rewrite left_absorb in H
  | H : context[ 1 × ?a ] |- _ => rewrite left_identity in H
  | H : context[ ?a × 1 ] |- _ => rewrite right_identity in H
  | H : 1 === 0 |- False => exact (field_not_trivial H)
  | H : ?a + - ?b === 0 |- _ => fold (a − b) in H
  | H : ?a − ?b === 0 |- _ => apply (field_zero_diff_eq _) in H
  | H : 0 === _ |- _ => apply symmetry in H
  end.

Opaque LC.sub.


Lemma lc_one_spec σ ω β (Hβ : solution ω σ β) :
  ω ⟦ lc_one fone ⟧ β === 1.
Proof.
  unfold lc_one. rewrite (LC.den_single F), right_identity.
  destruct (equiv_dec (β xH) 1 ) as [ EQ | NEQ ]; auto.
  destruct (not_none (proj1 Hβ  _ _ (proj2 (_.(md_invariant))))) as ( one & hone ).
  elim (project_inj hone project_one NEQ).
  apply project_in in hone. exact (Interval.const_exact _ _ hone).
  apply Pos.le_1_l.
Qed.


(* Properties about projection *)

Remark project_const_Some x y z:
  project x (const y) = Some z → z = y.
Proof.
  intro H. apply project_in in H. unfold denote, is_true in H. simpl in H. rewrite andb_true_iff in H. destruct H as [H1 H2]. apply Zle_bool_imp_le in H1; apply Zle_bool_imp_le in H2. Psatz.lia.
Qed.


(* Properties about input_bit *)

Lemma input_bit_spec (lc : Lc f) (σ σ' : cqap f) (Hincr : incr σ σ') :
  input_bit σ = STm.Result lc σ' Hincr →
  (∀ ω : LC.var, invariant ω σ) →
  ∀ (ω : LC.var),
    (∀ (ρ : valuation), solution ω σ' ρ → project (ω⟦ lc ⟧ ρ) Interval.bit ≠ None) ∧
    (invariant ω σ').
Proof.
  pose proof (solution_monotone Hincr) as hm. unfold input_bit, fresh_eqn, fresh_wire. intros H1 INV ω. hsplit. subst. simpl in *. subst. split.
  - intros β hβ.
    pose proof (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)) as K.
    assert (next_wire σ <= ω)%positive.
    generalize (proj2 (proj2 hβ)). simpl. clear. ppsimpl. Psatz.lia.
    unfold lc_of_wire in *.
    rewrite (LC.den_sub F), ! (LC.den_single F), (LC.den_zero F) in K
      by auto using Pos.le_1_l.
    rewrite (LC.den_single F) by auto.
    clean_field.
    apply (field_integral F) in K. rewrite right_identity.
    destruct K as [ K | K ].
    rewrite K, project_zero; intro; hsplit.
    apply (field_zero_diff_eq F) in K. rewrite <- K.
    rewrite lc_one_spec, project_one; auto. discriminate.
  - intros β hβ x. simpl in *.
    exact (INV _ _ (hm _ _ hβ) x).
Qed.

Lemma in_bit z : z ∈ bit → z = 0%Z ∨ z = 1%Z.
Proof. destruct z as [ |[p|p| ]|z]; auto; discriminate. Qed.
Arguments in_bit {_} _.


(* Properties about size *)

Lemma size_muln_2 : ∀ r, size (muln 2 r) = (2 * (size r))%N.
Proof.
  intros [r R]. unfold muln, size. simpl.
  replace (R*2 - r*2) with ((R - r)*2) by Psatz.lia.
  case (R - r); auto.
  intro p. replace (Z.pos p * 2) with (Zpos (p~0)) by Psatz.lia.
  reflexivity.
Qed.


(* Properties about the range of binary representation *)

Lemma range_of_bit_numbers_S q :
  range_of_bit_numbers (S q) =
  add bit (muln 2 (range_of_bit_numbers q)).
Proof.
  change (add bit (muln 2 (range_of_bit_numbers q))) with {|imin := 0; imax := 1 + (sup_of_bit_numbers q) * 2|}.
  unfold range_of_bit_numbers.
  replace (sup_of_bit_numbers (S q)) with (Z.add (Zpos xH) (Z.mul (sup_of_bit_numbers q) (Zpos (xO xH)))); auto.
  unfold sup_of_bit_numbers. rewrite two_power_nat_S. clear. Psatz.lia.
Qed.

Lemma size_range_of_bit_numbers q :
  size (range_of_bit_numbers q) = Z.to_N (sup_of_bit_numbers q).
Proof.
  unfold range_of_bit_numbers, sup_of_bit_numbers, size. simpl.
  now rewrite <- Zminus_0_l_reverse.
Qed.

Lemma size_range_of_binary_repr (bs:binary_repr f) :
  size (range_of_binary_repr bs) = Z.to_N (sup_of_bit_numbers (length bs)).
Proof. apply size_range_of_bit_numbers. Qed.

Lemma size_range_of_bit_numbers_smaller q q' :
  (q ≤ q') →
  (size (range_of_bit_numbers q) <= size (range_of_bit_numbers q'))%N.
Proof.
  clear. induction 1.
  - reflexivity.
  - apply (N.le_trans _ _ _ IHle).
    rewrite range_of_bit_numbers_S.
    unfold size, add, bit, muln; simpl.
    rewrite <- !Zminus_0_l_reverse.
    change ((fun x => (match x with | 0%Z => 0 | Z.pos p => N.pos p | Z.neg _ => 0 end <= match match (x * 2)%Z with | 0%Z => 1%Z | Z.pos y' => Z.pos match y' with | xI q0 => (Pos.succ q0)~0 | xO q0 => q0~1 | xH => 2 end | Z.neg y' => Z.pos_sub 1 y' end with | 0%Z => 0 | Z.pos p => N.pos p | Z.neg _ => 0 end)%N) (sup_of_bit_numbers m)).
    case (sup_of_bit_numbers m); simpl; intros; try Psatz.lia.
    replace (p * 2)%positive with (2 * p)%positive by Psatz.lia.
    simpl. Psatz.lia.
Qed.

Lemma size_range_of_bit_numbers_muln_S q :
  (size (muln 2 (range_of_bit_numbers q)) <= size (range_of_bit_numbers (S q)))%N.
Proof.
  clear.
  rewrite range_of_bit_numbers_S.
  unfold size, add, bit, muln; simpl.
  rewrite <- !Zminus_0_l_reverse.
  change ((fun x => (match x with | 0%Z => 0 | Z.pos p => N.pos p | Z.neg _ => 0 end <= match match x with | 0%Z => 1%Z | Z.pos y' => Z.pos match y' with | xI q0 => (Pos.succ q0)~0 | xO q0 => q0~1 | xH => 2 end | Z.neg y' => Z.pos_sub 1 y' end with | 0%Z => 0 | Z.pos p => N.pos p | Z.neg _ => 0 end)%N) (sup_of_bit_numbers q * 2)%Z).
  case (sup_of_bit_numbers q * 2)%Z; simpl; intros; try Psatz.lia.
  case p; intros; simpl; Psatz.lia.
Qed.

Lemma sup_of_bit_numbers_range (x: binary_repr f) :
  (size (range_of_binary_repr x) < Npos large_prime)%N →
  0 <= sup_of_bit_numbers (length x) < Zpos large_prime.
Proof.
  clear.
  unfold range_of_binary_repr. generalize (length x). clear x. intros n.
  unfold range_of_bit_numbers, sup_of_bit_numbers, size.
  rewrite <- Zminus_0_l_reverse. unfold imax.
  generalize (Coqlib.two_power_nat_pos n).
  destruct (two_power_nat _) as [ | p | p ]. Psatz.lia.
  intros _. destruct (Zpos p - 1) eqn:?; Psatz.lia.
  Psatz.lia.
Qed.

Lemma size_range_of_bit_numbers_wordsize x
  (H: (x ≤ Int.wordsize)) :
  (size (range_of_bit_numbers x) < N.pos large_prime)%N.
Proof.
  clear -H.
  apply (N.le_lt_trans _ (size (range_of_bit_numbers 32))); auto using size_range_of_bit_numbers_smaller.
  reflexivity.
Qed.

Lemma size_range_of_binary_repr_wordsize (bs:binary_repr f)
  (H: (length bs ≤ Int.wordsize)) :
  (size (range_of_binary_repr bs) < N.pos large_prime)%N.
Proof.
  unfold range_of_binary_repr.
  auto using size_range_of_bit_numbers_wordsize.
Qed.

(* Properties about binary representation *)

Lemma lc_of_binary_repr_spec ω σ (β : valuation) (hβ : solution ω σ β) bs :
  (size (range_of_binary_repr bs) < Npos large_prime)%N →
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
  project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) ≠ None.
Proof.
  induction bs as [ |b bs IHbs]; simpl.
  - intros _ _. rewrite LC.den_zero; eauto.
    rewrite <- f_of_z_zero. rewrite project_const. discriminate.
  - intros Hsup HFor. inversion_clear HFor.
    unfold range_of_binary_repr in Hsup; simpl in Hsup.
    assert (Hsup': (size (range_of_binary_repr bs) < N.pos large_prime)%N) by
        (unfold range_of_binary_repr; apply (N.le_lt_trans _ (size (range_of_bit_numbers (S (length bs))))); auto;
         apply size_range_of_bit_numbers_smaller; clear; Psatz.lia).
    assert (IH:=IHbs Hsup' H0). clear IHbs H0.
      rewrite LC.den_add, LC.den_muln; eauto.
      unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
      destruct (not_none H) as [p1 Hp1].
      destruct (not_none IH) as [p2 Hp2].
      rewrite (project_add (y':=(2 * p2)%Z) Hp1).
      * discriminate.
      * rewrite (project_muln _ Hp2); auto.
        apply (N.le_lt_trans _ (size (range_of_bit_numbers (S (length bs))))); auto.
        apply size_range_of_bit_numbers_muln_S.
      * rewrite <- range_of_bit_numbers_S. auto.
Qed.


(* Properties about fresh binary decompositions *)

Lemma fresh_binary_decomposition_spec_gen :
  ∀ n acc σ σ' (Hincr : incr σ σ') bs,
    fresh_binary_decomposition_aux n acc σ = STm.Result bs σ' Hincr →
    (∀ ω : LC.var, invariant ω σ) →
    ∀ (ω : LC.var), (invariant ω σ') ∧
      ∀ (β : valuation) (hβ : solution ω σ' β),
        (List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) acc →
         List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs).
Proof.
  induction n as [ |n IHn].
  - simpl. intros acc σ σ' Hincr bs H INV ω. hsplit. subst. simpl in *. split; auto.
  - simpl. intros acc σ σ' Hincr bs H INV ω. hsplit. subst.
    assert (H2:=input_bit_spec H0 INV); clear H0. rewrite and_forall_comm in H2. destruct H2 as [H2 H3].
    destruct (IHn _ _ _ _ _ H H3 ω) as [IH1 IH2]; clear IHn. split; auto.
    intros β hβ Hacc. apply IH2; clear IH2; auto.
    constructor; auto.
    apply H2. auto using (solution_monotone Hincr'').
Qed.


Lemma fresh_binary_decomposition_spec :
  ∀ n σ σ' (Hincr : incr σ σ') bs,
    fresh_binary_decomposition n σ = STm.Result bs σ' Hincr →
    (∀ ω : LC.var, invariant ω σ) →
    ∀ (ω : LC.var), (invariant ω σ') ∧
      ∀ (β : valuation) (hβ : solution ω σ' β),
        (List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs).
Proof.
  intros n σ σ' Hincr bs Hbs INV ω.
  unfold fresh_binary_decomposition in Hbs.
  hsplit. repeat subst. simpl in *.
  case_eq ((size (range_of_binary_repr x) <? N.pos large_prime)%N); intro Heq; rewrite Heq in Hbs; try discriminate.
  hsplit. repeat subst. simpl in *.
  destruct (fresh_binary_decomposition_spec_gen n nil Hbs0 INV ω) as [H1 H2]; eauto.
Qed.


Lemma positive_binary_decomposition_aux_spec ω σ β (hβ: solution ω σ β) p : ∀ n, (n ≤ Int.wordsize) →
  let bs := positive_binary_decomposition_aux p n in
  (length bs ≤ n) ∧
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
  ∃ p',
    project (ω ⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p' ∧
    ∃ k : Z, Zpos p = (k * two_power_nat n + p')%Z.
Proof.
  induction p as [p IHp|p IHp| ].
  - intros [ |n]; simpl.
    * repeat split; auto.
      exists 0%Z. rewrite (LC.den_zero F), <- f_of_z_zero, project_const.
      split; auto.
      exists (Z.pos p~1). change (two_power_nat 0) with 1%Z. Psatz.lia.
    * intro Hn. assert (Hn': n ≤ Int.wordsize) by Psatz.lia.
      destruct (IHp n Hn') as (IH1 & IH2 & p' & IH3 & k & IH4).
      split; [Psatz.lia| ].
      split; [constructor; auto; rewrite lc_one_spec, project_one; eauto; discriminate| ].
      exists (1 + 2*p')%Z. split.
      + unfold range_of_binary_repr. simpl.
        rewrite range_of_bit_numbers_S.
        rewrite (LC.den_add F), (LC.den_muln F), (lc_one_spec hβ).
        rewrite (project_add (x':=1%Z) (y':=2*p')).
        reflexivity.
        apply project_one.
        rewrite (project_muln _ IH3).
        reflexivity.
        clear -Hn' IH1.
        unfold range_of_binary_repr.
        apply (N.le_lt_trans _ (size (muln 2 (range_of_bit_numbers 32)))); [ |reflexivity].
        rewrite !size_muln_2.
        assert ((size (range_of_bit_numbers (length (positive_binary_decomposition_aux p n))) <= size (range_of_bit_numbers 32))%N); [ |Psatz.lia].
        apply size_range_of_bit_numbers_smaller.
        eauto using le_trans.
        rewrite <- range_of_bit_numbers_S.
        apply (N.le_lt_trans _ (size (range_of_bit_numbers 33))); [ |reflexivity].
        apply size_range_of_bit_numbers_smaller.
        clear - Hn' IH1. compute in Hn'. Psatz.lia.
      + exists k. clear -IH4. rewrite two_power_nat_S. Psatz.lia.
  - intros [ |n]; simpl.
    * repeat split; auto.
      exists 0%Z. rewrite (LC.den_zero F), <- f_of_z_zero, project_const.
      split; auto.
      exists (Z.pos p~0). change (two_power_nat 0) with 1%Z. Psatz.lia.
    * intro Hn. assert (Hn': n ≤ Int.wordsize) by Psatz.lia.
      destruct (IHp n Hn') as (IH1 & IH2 & p' & IH3 & k & IH4).
      split; [Psatz.lia| ].
      split; [constructor; auto; rewrite (LC.den_zero F), project_zero; eauto; discriminate| ].
      exists (2*p')%Z. split.
      + unfold range_of_binary_repr. simpl.
        rewrite range_of_bit_numbers_S.
        rewrite (LC.den_add F), (LC.den_muln F), (LC.den_zero F).
        rewrite (project_add (x':=0%Z) (y':=2*p')).
        reflexivity.
        apply project_zero.
        rewrite (project_muln _ IH3).
        reflexivity.
        clear -Hn' IH1.
        unfold range_of_binary_repr.
        apply (N.le_lt_trans _ (size (muln 2 (range_of_bit_numbers 32)))); [ |reflexivity].
        rewrite !size_muln_2.
        assert ((size (range_of_bit_numbers (length (positive_binary_decomposition_aux p n))) <= size (range_of_bit_numbers 32))%N); [ |Psatz.lia].
        apply size_range_of_bit_numbers_smaller.
        eauto using le_trans.
        rewrite <- range_of_bit_numbers_S.
        apply (N.le_lt_trans _ (size (range_of_bit_numbers 33))); [ |reflexivity].
        apply size_range_of_bit_numbers_smaller.
        clear - Hn' IH1. compute in Hn'. Psatz.lia.
      + exists k. clear -IH4. rewrite two_power_nat_S. Psatz.lia.
  - intros [ |n]; simpl.
    * repeat split; auto.
      exists 0%Z. rewrite (LC.den_zero F), <- f_of_z_zero, project_const.
      split; auto.
      exists 1%Z. change (two_power_nat 0) with 1%Z. Psatz.lia.
    * intro Hn. split; [Psatz.lia| ].
      split; [repeat constructor; rewrite (lc_one_spec hβ), project_one; discriminate| ].
      exists 1%Z. split.
      + rewrite (LC.den_add F), (LC.den_muln F), (LC.den_zero F), (lc_one_spec hβ). assert (H: 1 + f_of_z 2 × 0 === 1) by ring. rewrite H. apply project_one.
      + exists 0%Z. Psatz.lia.
Qed.


Lemma positive_binary_decomposition_spec ω σ β (hβ: solution ω σ β) p :
  let bs := positive_binary_decomposition p in
  (length bs <= Int.wordsize)%nat ∧
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
  ∃ p',
    project (ω ⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p' ∧
    Int.repr p' = Int.repr (Zpos p).
Proof.
  destruct (positive_binary_decomposition_aux_spec hβ p (le_n Int.wordsize)) as (H1 & H2 & p' & H3 & k & H4).
  unfold positive_binary_decomposition. repeat split; auto.
  exists p'. split; auto.
  symmetry. apply Int.eqm_samerepr.
  now exists k.
Qed.


Lemma fresh_binary_decomposition_aux_len n :
  ∀ (acc: binary_repr f) s x s' H,
    fresh_binary_decomposition_aux n acc s = STm.Result x s' H →
    (length acc + n = length x)%nat.
Proof.
  clear.
  unfold fresh_binary_decomposition_aux.
  induction n as [ | n IH ]; intros acc s x s' hIncr H;
  hsplit; subst. Psatz.lia.
  specialize (IH _ _ _ _ _ H). simpl in IH. Psatz.lia.
Qed.

(* Properties about truncation *)

Section Truncate.

  Variables (ω : LC.var) (σ : cqap f) (β : valuation) (Hβ : solution ω σ β).

  Lemma project_nil :
    project (ω ⟦ lc_of_binary_repr nil ⟧ β) (@range_of_binary_repr f nil) = Some 0%Z.
  Proof.
    rewrite LC.den_zero; eauto.
    now rewrite <- f_of_z_zero, project_const.
  Qed.

  Lemma project_cons b bs x y
        (Hb: project (ω ⟦ b ⟧ β) bit = Some x)
        (Hbs : project (ω ⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some y)
        (Hsize : (size (range_of_binary_repr (b::bs)%list) < N.pos large_prime)%N) :
    project (ω ⟦ lc_of_binary_repr (b::bs)%list ⟧ β) (range_of_binary_repr (b::bs)%list) = Some (x + 2 * y)%Z.
  Proof.
    unfold range_of_binary_repr in *. simpl.
    rewrite range_of_bit_numbers_S.
    rewrite LC.den_add, LC.den_muln; eauto.
    rewrite (project_add (y':=2 * y) Hb); auto.
    - rewrite (project_muln 2 Hbs); auto.
      apply (N.le_lt_trans _ (size (range_of_bit_numbers (length (b :: bs))))); auto.
      simpl. apply size_range_of_bit_numbers_muln_S.
    - rewrite <- range_of_bit_numbers_S; now auto.
  Qed.

  Section GenStart.

    Variable start : nat → binary_repr f.
    Hypothesis start_zero : ∀ n,
      List.Forall (fun b => b = LC.zero f) (start n).
    Hypothesis start_zero_nil : start 0%nat = nil.
    Hypothesis start_size : ∀ n m, n ≤ m →
      (size (range_of_binary_repr (start n)) <= size (range_of_binary_repr (start m)))%N.


    Lemma start_bit n:
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) (start n).
    Proof.
      generalize (start_zero n). rewrite !List.Forall_forall.
      intros sz b Hin.
      rewrite (sz _ Hin). rewrite LC.den_zero; eauto.
      rewrite project_zero. discriminate.
    Qed.

    Lemma start_den_zero n :
      (size (range_of_binary_repr (start n)) < N.pos large_prime)%N →
      project (ω ⟦ lc_of_binary_repr (start n) ⟧ β)
              (range_of_binary_repr (start n)) = Some 0%Z.
    Proof.
      assert (H: ∀ bs, (size (range_of_binary_repr bs) < N.pos large_prime)%N → List.Forall (fun b => b = LC.zero f) bs → project (ω ⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some 0%Z); auto.
      induction bs as [ |b bs IHbs].
      - intros _ _; apply project_nil.
      - intros Hsize H. inversion H. subst. clear H.
        rewrite (project_cons (x:=0%Z) (y:=0%Z)); auto.
        * rewrite LC.den_zero; eauto.
        * apply IHbs; auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (LC.zero f :: bs)%list))); auto.
          unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_smaller; auto.
    Qed.

    Lemma truncate_length bs : ∀ n,
      length (truncate n start bs) =
      if Nat.leb n (length bs) then n else
        ((length bs) + (length (start (n - (length bs)))))%nat.
    Proof.
      induction bs as [ |b bs IHbs]; simpl.
      - intro n. case_eq (Nat.leb n 0); intro Heq.
        * rewrite Nat.leb_le in Heq. assert (H1:=le_n_0_eq _ Heq). subst n. now rewrite start_zero_nil.
        * now replace (n - 0)%nat with n by Psatz.lia.
      - intros [ |n]; simpl; auto.
        rewrite IHbs. now case (Nat.leb n (length bs)).
    Qed.

    Lemma truncate_spec bs : ∀ n,
      let bs' := truncate n start bs in
      (size (range_of_binary_repr bs) < N.pos large_prime)%N →
      (size (range_of_binary_repr bs') < N.pos large_prime)%N →
      (size (range_of_binary_repr (start n)) < N.pos large_prime)%N →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
                  (∀ p p',
                     project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p →
                     project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p' →
                   ∃ k : Z, p = (k * two_power_nat n + p')%Z).
    Proof.
      induction bs as [ |b bs IHbs]; simpl.
      - intros n _ _ Hsizen; split.
        * apply start_bit.
        * intros p p'. rewrite LC.den_zero, <- (f_of_z_zero), project_const by eauto.
          intros H1 H2; inversion_clear H1.
          rewrite start_den_zero in H2; auto. inversion_clear H2.
          exists 0%Z. Psatz.lia.
      - intros [ |n] Hsize1 Hsize2 Hsizen HF.
        * split.
          + constructor.
          + intros p p' Hp. rewrite LC.den_zero, <- (f_of_z_zero), project_const by eauto.
            intro Hp'; inversion_clear Hp'.
            exists p. change (two_power_nat 0) with 1%Z. Psatz.lia.
        * inversion HF; subst; clear HF.
          assert (H10:(size (range_of_binary_repr bs) < N.pos large_prime)%N) by (apply (N.le_lt_trans _ (size (range_of_binary_repr (b :: bs)%list))); auto; unfold range_of_binary_repr; simpl; apply size_range_of_bit_numbers_smaller; auto).
          assert (H11:(size (range_of_binary_repr (truncate n start bs)) < N.pos large_prime)%N) by (apply (N.le_lt_trans _ (size (range_of_binary_repr (b :: (truncate n start bs))%list))); auto; unfold range_of_binary_repr; simpl; apply size_range_of_bit_numbers_smaller; auto).
          assert (H13:(size (range_of_binary_repr (start n)) < N.pos large_prime)%N) by (apply (N.le_lt_trans _ (size (range_of_binary_repr (start (S n))))); auto).
          generalize (IHbs n H10 H11 H13 H2); clear IHbs. simpl. intros [IH1 IH2]. split.
          + constructor; auto.
          + intros p p'.
            destruct (not_none H1) as [b0 Hb].
            destruct (not_none (lc_of_binary_repr_spec Hβ H10 H2)) as [z Hz].
            destruct (not_none (lc_of_binary_repr_spec Hβ H11 IH1)) as [z' Hz'].
            destruct (IH2 _ _ Hz Hz') as [k Hk].
            rewrite !LC.den_add, !LC.den_muln; eauto.
            unfold range_of_binary_repr; simpl.
            rewrite !range_of_bit_numbers_S.
            rewrite (project_add (y':= 2*z) Hb).
            intro H12. assert (p = (b0+2*z)%Z) by (inversion H12; auto). subst p; clear H12.
            rewrite (project_add (y':= 2*z') Hb).
            intro H12. assert (p' = (b0+2*z')%Z) by (inversion H12; auto). subst p'; clear H12.
            exists k. subst z. rewrite two_power_nat_S. clear. Psatz.lia.
            rewrite (project_muln _ Hz'); auto.
            apply (N.le_lt_trans _ (size (range_of_binary_repr (b :: truncate n start bs)%list))); auto.
            unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_muln_S.
            rewrite <- range_of_bit_numbers_S. exact Hsize2.
            rewrite (project_muln _ Hz); auto.
            apply (N.le_lt_trans _ (size (range_of_binary_repr (b :: bs)%list))); auto.
            unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_muln_S.
            rewrite <- range_of_bit_numbers_S. exact Hsize1.
    Qed.

  End GenStart.


  Section Truncate32.

    Let start : nat → binary_repr f := fun _ => nil.

    Let start_zero n :
      List.Forall (fun b => b = LC.zero f) (start n).
    Proof. constructor. Qed.

    Let start_zero_nil : start 0%nat = nil.
    Proof. reflexivity. Qed.

    Let start_size n m : n ≤ m →
      (size (range_of_binary_repr (start n)) <= size (range_of_binary_repr (start m)))%N.
    Proof. reflexivity. Qed.

    Lemma truncate_32_length bs :
      (length (@truncate_32 f bs) <= Int.wordsize)%nat.
    Proof.
      unfold truncate_32. rewrite truncate_length.
      case_eq (Nat.leb Int.wordsize (length bs)); auto.
      rewrite NPeano_leb_lt, plus_0_r. auto with arith.
      reflexivity.
    Qed.

    Lemma size_range_of_binary_repr_truncate_32 bs :
      (size (range_of_binary_repr (@truncate_32 f bs)) < N.pos large_prime)%N.
    Proof.
      auto using size_range_of_binary_repr_wordsize, truncate_32_length.
    Qed.

    Lemma size_start_32 :
      (size (range_of_binary_repr (start Int.wordsize)) < N.pos large_prime)%N.
    Proof. reflexivity. Qed.

    Lemma truncate_32_spec bs :
      let bs' := truncate_32 bs in
      (size (range_of_binary_repr bs) < N.pos large_prime)%N →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
                  (∃ p p',
                     project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
                     project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p' ∧
                     Int.repr p = Int.repr p').
    Proof.
      intros bs' Hsize HFor. subst bs'.
      destruct (truncate_spec _ start_zero start_size (bs:=bs) Int.wordsize Hsize (size_range_of_binary_repr_truncate_32 bs) size_start_32 HFor) as [H1 H2]. split; auto.
      destruct (not_none (lc_of_binary_repr_spec Hβ Hsize HFor)) as [p Hp].
      destruct (not_none (lc_of_binary_repr_spec Hβ (size_range_of_binary_repr_truncate_32 bs) H1)) as [p' Hp'].
      exists p, p'. repeat split; auto.
      apply Int.eqm_samerepr, H2; assumption.
    Qed.

  End Truncate32.


  Section TruncateComplete32.

    Let start : nat → binary_repr f := zeros f.

    Let start_zero n :
      List.Forall (fun b => b = LC.zero f) (start n).
    Proof. induction n as [ |n IHn]; simpl; auto. Qed.

    Let start_zero_nil : start 0%nat = nil.
    Proof. reflexivity. Qed.

    Lemma start_length n : length (start n) = n.
    Proof. induction n as [ |n IHn]; simpl; auto. Qed.

    Let start_size n m (H:n ≤ m) :
      (size (range_of_binary_repr (start n)) <= size (range_of_binary_repr (start m)))%N.
    Proof.
      unfold range_of_binary_repr. rewrite !start_length.
      now apply size_range_of_bit_numbers_smaller.
    Qed.

    Lemma truncate_and_complete_32_length_eq bs :
      (length (@truncate_and_complete_32 f bs) = 32)%nat.
    Proof.
      unfold truncate_and_complete_32. rewrite truncate_length.
      case_eq (Nat.leb Int.wordsize (length bs)); auto.
      rewrite NPeano_leb_lt. rewrite start_length.
      intro H. rewrite le_plus_minus_r; auto with arith.
      reflexivity.
    Qed.

    Lemma truncate_and_complete_32_length bs :
      (length (@truncate_and_complete_32 f bs) ≤ 32)%nat.
    Proof. now rewrite truncate_and_complete_32_length_eq. Qed.

    Lemma size_range_of_binary_repr_truncate_and_complete_32 bs :
      (size (range_of_binary_repr (@truncate_and_complete_32 f bs)) < N.pos large_prime)%N.
    Proof.
      auto using size_range_of_binary_repr_wordsize, truncate_and_complete_32_length.
    Qed.

    Lemma size_zeros_32 :
      (size (range_of_binary_repr (start Int.wordsize)) < N.pos large_prime)%N.
    Proof. reflexivity. Qed.

    Lemma truncate_and_complete_32_spec bs :
      let bs' := truncate_and_complete_32 bs in
      (size (range_of_binary_repr bs) < N.pos large_prime)%N →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
                  (∃ p p',
                     project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
                     project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p' ∧
                     Int.repr p = Int.repr p').
    Proof.
      intros bs' Hsize HFor. subst bs'.
      destruct (truncate_spec _ start_zero start_size (bs:=bs) Int.wordsize Hsize (size_range_of_binary_repr_truncate_and_complete_32 bs) size_zeros_32 HFor) as [H1 H2]. split; auto.
      destruct (not_none (lc_of_binary_repr_spec Hβ Hsize HFor)) as [p Hp].
      destruct (not_none (lc_of_binary_repr_spec Hβ (size_range_of_binary_repr_truncate_and_complete_32 bs) H1)) as [p' Hp'].
      exists p, p'. repeat split; auto.
      apply Int.eqm_samerepr, H2; assumption.
    Qed.

  End TruncateComplete32.

End Truncate.

Lemma sup_of_bit_number_of_bits r :
  imax r <=? sup_of_bit_numbers (number_of_bits r).
Proof.
clear. unfold number_of_bits, sup_of_bit_numbers.
destruct r as [ x r ]. unfold imax. clear x.
rewrite two_power_nat_equiv.
apply Z.leb_le.
rewrite Nat2Z.inj_succ.
rewrite Z2Nat.id. 2: apply Z.log2_nonneg.
rewrite Z.pow_succ_r. 2: apply Z.log2_nonneg.
destruct r as [ | p | p ]. simpl; Psatz.lia. 2: simpl; Psatz.lia.
generalize (Z.log2_spec (Zpos p) (Pos2Z.is_pos p)).
rewrite Z.pow_succ_r. 2: apply Z.log2_nonneg.
Psatz.lia.
Qed.

(* Properties about get_binary_decomposition *)

Lemma get_binary_decomposition_spec :
  ∀ a σ σ' (Hincr : incr σ σ') bs hint,
    get_binary_decomposition a σ = STm.Result (bs, hint) σ' Hincr →
    (∀ ω : LC.var, invariant ω σ) →
    ∀ (ω : LC.var), (invariant ω σ') ∧
      ∀ (β : valuation) (hβ : solution ω σ' β),
      (length bs <= Int.wordsize)%nat ∧
      (List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs) ∧
      (exists lc Hincr',
          (get_lc a σ = STm.Result lc σ Hincr') ∧
          (exists p p',
             (project (ω⟦ fst lc ⟧ β) (snd lc) = Some p) ∧
             (project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p') ∧
             Int.repr p = Int.repr p')).
Proof.
  Opaque number_of_bits.
  unfold QAP.get_binary_decomposition.
  intros a σ σ' Hincr bs hint Hbs INV.
  pose proof (solution_monotone Hincr) as hm.
  hsplit. repeat subst. simpl in *.
  case_eq ((bcache s'') ! a); [intro b| ]; intro Hb; rewrite Hb in Hbs.
  - hsplit. repeat subst. simpl in *. intro ω. split; auto.
    intros β hβ. destruct (INV ω β hβ a) as (H1 & H2 & H20).
    destruct (H2 _ Hb) as (H300 & H3 & val & p & p' & H4 & H5 & H6 & H7).
    split; auto.
    split; auto.
    exists val, (incr_trans (incr_refl σ') (incr_refl σ')).
    unfold QAP.get_lc. hsplit. simpl.
    rewrite get_env_Some in H4. destruct H4 as [val1 [K1 K2]].
    unfold STm.bind. rewrite K1, K2. simpl in *.
    repeat split; auto.
    exists p, p'. repeat split; auto.
 - case_eq ((env s'') ! a); [intro val| ]; intro Hval; rewrite Hval in Hbs; [ |hsplit].
    case_eq val; [intros [lc r]| intro bs']; intro Heq; rewrite Heq in Hbs.
    * case_eq (is_const r); [intros [ |p|p]| ]; intro Hp; rewrite Hp in Hbs.
      + hsplit. repeat subst. simpl in *. intro ω. split; auto.
        intros β hβ; split; [Psatz.lia| ].
        split; auto.
        destruct (INV _ _ hβ a) as (H10 & H11 & H12).
        exists (lc, r), (incr_trans (incr_refl σ') (incr_refl σ')).
        unfold QAP.get_lc, STm.bind. simpl. rewrite Hval. simpl.
        split; auto.
        exists 0%Z, 0%Z.
        generalize (H10 (lc, r)). unfold get_env. rewrite Hval. simpl. intro H13. destruct (H13 (refl_equal _)) as (p & Hp1 & Hp2).
        rewrite Hp1. apply project_in in Hp1.
        destruct (is_const_elim _ Hp) as (H14 & H15).
        apply H15 in Hp1. subst p. repeat split; auto.
        rewrite LC.den_zero; eauto.
        unfold range_of_binary_repr, range_of_bit_numbers, sup_of_bit_numbers. simpl. rewrite <- f_of_z_zero. apply project_const.
      + unfold cache_binary_decomposition in Hbs.
        hsplit. repeat subst. simpl in *. split.
        { intros β hβ v. simpl.
          destruct (positive_binary_decomposition_spec hβ p) as (H300 & H10 & p' & H14 & H114).
          destruct (INV _ _ (hm _ _ hβ) v) as (H11 & H12 & H13).
          split; [ |split]; auto.
          intro bs0. rewrite M.gsspec. case (Coqlib.peq v a); intro EQ.

          subst v. intro H; inversion H; subst bs0; clear H.
          split; auto.
          split; auto.
          generalize (H11 (lc, r)). unfold get_env. rewrite Hval. simpl. intro H15. destruct (H15 (refl_equal _)) as (p0 & Hp0 & Hp0'). clear H15.
          exists (lc, r), p0, p'. simpl. repeat split; auto.
          apply project_in in Hp0.
          destruct (is_const_elim _ Hp) as (H17 & H18).
          apply H18 in Hp0. subst p0. auto.

          apply H12.

          intro H15. rewrite M.gso; auto.
          intro H. subst v. rewrite (proj2 (wf s'') _ H15) in Hval. discriminate. }
        { intros β hβ.
          destruct (positive_binary_decomposition_spec hβ p) as (H300 & H10 & p' & H14 & H114).
          destruct (INV _ _ (hm _ _ hβ) a) as (H11 & H12 & H13).
          split; auto.
          split; auto.
          exists (lc, r), (incr_trans (incr_refl s'') (incr_refl s'')).
          unfold QAP.get_lc, STm.bind. simpl. rewrite Hval. simpl. split; auto.
          generalize (H11 (lc, r)). unfold get_env. rewrite Hval. simpl. intro H15. destruct (H15 (refl_equal _)) as (p0 & Hp0 & Hp0'). clear H15.
          exists p0, p'. repeat split; auto.
          apply project_in in Hp0.
          destruct (is_const_elim _ Hp) as (H17 & H18).
          apply H18 in Hp0. subst p0. auto. }
      + hsplit.
      (* We have to do a fresh BD *)
      + hsplit.
        pose proof (solution_monotone Hincr'') as hm''0.
        repeat subst. simpl in *. intro ω.
        destruct (fresh_binary_decomposition_spec _ Hbs0 INV ω) as (INV2 & H10).
        unfold fresh_eqn in Hbs2. unfold cache_binary_decomposition in Hbs4. hsplit. repeat subst. simpl in *.
        split.
        { intros β hβ v. simpl.
          destruct (INV2 β (hm''0 _ _ hβ) v) as (H11 & H12 & H13).
          split; auto. split.

          intro bs. rewrite M.gsspec. case (Coqlib.peq v a); intro EQ.
          subst v. intro H; inversion H; subst bs; clear H.
          assert (H14:(size (range_of_binary_repr x) < N.pos large_prime)%N).
          unfold fresh_binary_decomposition in Hbs0.
          hsplit. repeat subst. simpl in *.
          case_eq ((size (range_of_binary_repr x0) <? N.pos large_prime)%N); intro HE; rewrite HE in Hbs0; hsplit.
          repeat subst. simpl in *. rewrite N.ltb_lt in HE. auto.
          destruct (truncate_32_spec (hm''0 _ _ hβ) H14 (H10 _ (hm''0 _ _ hβ))) as (H15 & p & p' & H16 & H17 & H18).
          split; auto using truncate_32_length.
          split; auto.
          exists (lc, r). unfold get_env.
          destruct Hincr' as (H300 & H301 & H302 & H303 & H304 & H305 & H306 & H20).
          rewrite H20 in Hval.
          rewrite Hval. simpl.
          destruct (INV2 _ (hm''0 _ _ hβ) a) as (H21 & H22 & H23).
          generalize (H21 (lc, r)). unfold get_env. rewrite Hval. simpl. intro H24. destruct (H24 (refl_equal _)) as (p0 & Hp0 & Hp0'); clear H24.
          exists p0, p'; repeat split; auto.
          rewrite <- H18.
          destruct hβ as (_ & H25 & _). simpl in H25.
          assert (H26: (next_root s'' < Pos.succ (next_root s''))%positive) by (clear; Psatz.lia).
          generalize (H25 (next_root s'') H26 (LC.zero f) (LC.zero f) (LC.sub lc (lc_of_binary_repr x))).
          rewrite !M.gss. intro H27. specialize (H27 (refl_equal _) (refl_equal _) (refl_equal _)).
          rewrite (LC.den_zero F), (LC.den_sub F) in H27. ring_simplify in H27. symmetry in H27. apply (field_zero_diff_eq F) in H27.
          rewrite H27 in Hp0.
          f_equal.
          eapply project_subrange; try eassumption.
          unfold leb. simpl. rewrite e. simpl.
          replace (length x) with (number_of_bits r).
          apply sup_of_bit_number_of_bits.
          clear -Hbs0.
          unfold fresh_binary_decomposition in Hbs0. hsplit.
          flatten_eq hsplit. simpl in *. subst. simpl in *.
          apply (fresh_binary_decomposition_aux_len _ _ Hbs1).
          apply sup_of_bit_numbers_range; eassumption.
          destruct (wf s''0) as (_ & H21). specialize (H21 a).
          assert (H22: ((a < next_var s''0)%positive) ∨ ((next_var s''0 <= a)%positive)) by (clear; Psatz.lia).
          destruct H22 as [H22|H22]; auto.
          rewrite (H21 H22) in Hval. discriminate.
          apply H12.
          intro Hv. rewrite M.gso; auto.
          intro H. subst v.
          destruct Hincr' as (H300 & H301 & H302 & H303 & H304 & H305 & H306 & H20).
          rewrite H20 in Hval.
          destruct (wf s'') as (H22 & H21). rewrite (H21 a) in Hval; auto. discriminate.
          destruct (wf s''0) as (_ & H21). specialize (H21 a).
          assert (H22: ((a < next_var s''0)%positive) ∨ ((next_var s''0 <= a)%positive)) by (clear; Psatz.lia).
          destruct H22 as [H22|H22]; auto.
          rewrite (H21 H22) in Hval. discriminate. }
        { intros β hβ.
          assert (H14:(size (range_of_binary_repr x) < N.pos large_prime)%N).
          unfold fresh_binary_decomposition in Hbs0.
          hsplit. repeat subst. simpl in *.
          case_eq ((size (range_of_binary_repr x0) <? N.pos large_prime)%N); intro HE; rewrite HE in Hbs0; hsplit.
          repeat subst. simpl in *. rewrite N.ltb_lt in HE. auto.
          destruct (truncate_32_spec (hm''0 _ _ hβ) H14 (H10 _ (hm''0 _ _ hβ))) as (H15 & p & p' & H16 & H17 & H18).
          split; auto using truncate_32_length.
          split; auto.
          exists (lc, r), (incr_trans (incr_refl s''0) (incr_refl s''0)).
          unfold QAP.get_lc. simpl. unfold STm.bind, STm.get, STm.ret. rewrite Hval. simpl. split; auto.
          destruct (INV _ _ (hm _ _ hβ) a) as (H21 & H22 & H23).
          generalize (H21 (lc, r)). unfold get_env. rewrite Hval. simpl. intro H24. destruct (H24 (refl_equal _)) as (p0 & Hp0 & Hp0'); clear H24.
          exists p0, p'; repeat split; auto.
          rewrite <- H18.
          destruct hβ as (_ & H25 & _). simpl in H25.
          assert (H26: (next_root s'' < Pos.succ (next_root s''))%positive) by (clear; Psatz.lia).
          generalize (H25 (next_root s'') H26 (LC.zero f) (LC.zero f) (LC.sub lc (lc_of_binary_repr x))).
          rewrite !M.gss. intro H27. specialize (H27 (refl_equal _) (refl_equal _) (refl_equal _)).
          rewrite (LC.den_zero F), (LC.den_sub F) in H27. ring_simplify in H27. symmetry in H27. apply (field_zero_diff_eq F) in H27.
          rewrite H27 in Hp0.
          f_equal.
          eapply project_subrange; try eassumption.
          unfold leb. simpl. rewrite e. simpl.
          replace (length x) with (number_of_bits r).
          apply sup_of_bit_number_of_bits.
          clear -Hbs0.
          unfold fresh_binary_decomposition in Hbs0. hsplit.
          flatten_eq hsplit. simpl in *. subst. simpl in *.
          apply (fresh_binary_decomposition_aux_len _ _ Hbs1).
          apply sup_of_bit_numbers_range; eassumption.
        }
    * hsplit. repeat subst. simpl in *.
      intro ω.
      split; auto.
      intros β hβ.
      destruct (INV ω _ hβ a) as (H10 & H11 & H12).
      generalize (H10 (lc_of_binary_repr bs, range_of_binary_repr bs)).
      unfold get_env. rewrite Hval. simpl. intro H13.
      specialize (H13 refl_equal). destruct H13 as (p & H13 & H14).
      specialize (H14 _ refl_equal). destruct H14 as (H300 & H14).
      split; auto.
      split; auto.
      exists (lc_of_binary_repr bs, range_of_binary_repr bs), (incr_trans (incr_refl σ') (incr_refl σ')).
      unfold QAP.get_lc. simpl. unfold STm.bind. simpl. rewrite Hval. simpl. split; auto.
      exists p, p; repeat split; auto.
  Transparent number_of_bits.
Qed.


(* Properties about bitwise operations *)

Lemma bitwise_first_nil_length op : ∀ bs bs' (σ: cqap f) σ' Hincr,
  bitwise_first_nil op bs σ = STm.Result bs' σ' Hincr →
  length bs = length bs'.
Proof.
  induction bs as [ |lc2 bs IHbs]; simpl; intros bs' σ σ' Hincr H.
  - hsplit. repeat subst. reflexivity.
  - hsplit. repeat subst. simpl in *.
    now rewrite (IHbs _ _ _ _ H2).
Qed.


Lemma bitwise_aux_length op : ∀ bs1 bs2 bs (σ: cqap f) σ' Hincr,
  bitwise_aux op bs1 bs2 σ = STm.Result bs σ' Hincr →
  length bs = max (length bs1) (length bs2).
Proof.
  induction bs1 as [ |lc1 bs1 IHbs1]; simpl; intros bs2 bs σ σ' Hincr H.
  - now rewrite (bitwise_first_nil_length _ _ H).
  - destruct bs2 as [ |lc2 bs2]; simpl in *.
    * hsplit. repeat subst. simpl in *.
      rewrite (IHbs1 _ _ _ _ _ H2). simpl. now rewrite Max.max_0_r.
    * hsplit. repeat subst. simpl in *.
      now rewrite (IHbs1 _ _ _ _ _ H2).
Qed.


(* TODO: make the following more modular in the bitwise operation *)
(** Correctness of && **)

Definition bit_Z_land p1 p2 :=
  match p1, p2 with
  | 0%Z, 0%Z
  | 1%Z, 0%Z
  | 0%Z, 1%Z => Some 0%Z
  | 1%Z, 1%Z => Some 1%Z
  | _, _ => None
  end.

Lemma bitAnd_spec b1 b2 b σ σ' Hincr :
  bitAnd b1 b2 σ = STm.Result b σ' Hincr →
  (∀ ω, invariant ω σ) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      ∀ p1 p2,
      project (ω⟦ b1 ⟧ β) Interval.bit = Some p1 →
      project (ω⟦ b2 ⟧ β) Interval.bit = Some p2 →
      ∃ p, project (ω⟦ b ⟧ β) Interval.bit = Some p ∧
           bit_Z_land p1 p2 = Some p.
Proof.
  unfold bitAnd, fresh_wire, fresh_eqn.
  pose proof (solution_monotone Hincr) as hm.
  hsplit. repeat subst. simpl in *.
  case_eq (LC.is_single 1%positive b1); [intro d1| ]; intro Heq1.
  - intros H INV ω. hsplit. repeat subst. simpl in *.
    split; auto.
    intros β hβ p1 p2 Hp1 Hp2.
    assert (H10: (1%positive <= ω)%positive) by (clear; Psatz.lia).
    generalize (LC.den_is_single F β H10 Heq1).
    assert (H200: β 1%positive === 1).
      destruct (equiv_dec (β xH) 1 ) as [ EQ | NEQ ]; auto.
      destruct (not_none (proj1 hβ  _ _ (proj2 (_.(md_invariant))))) as ( one & hone ).
      elim (project_inj hone project_one NEQ).
      apply project_in in hone. exact (Interval.const_exact _ _ hone).
    rewrite H200.
    rewrite left_identity. intro H. rewrite H in Hp1.
    assert (Hp1' := project_in Hp1). assert (Hp2' := project_in Hp2).
    destruct (in_bit Hp1') as [H11|H11]; destruct (in_bit Hp2') as [H12|H12];
      subst p1; subst p2; clear Hp1'; clear Hp2';
      [exists 0%Z|exists 0%Z|exists 0%Z|exists 1%Z]; split; auto;
      rewrite (LC.den_muln F);
      change bit with (mul bit bit); rewrite (project_mul Hp1 Hp2); auto;
      reflexivity.
  - case_eq (LC.is_single 1%positive b2); [intro d2| ]; intro Heq2.
    * intros H INV ω. hsplit. repeat subst. simpl in *.
      split; auto.
      intros β hβ p1 p2 Hp1 Hp2.
      assert (H10: (1%positive <= ω)%positive) by (clear; Psatz.lia).
      generalize (LC.den_is_single F β H10 Heq2).
      assert (H200: β 1%positive === 1).
        destruct (equiv_dec (β xH) 1 ) as [ EQ | NEQ ]; auto.
        destruct (not_none (proj1 hβ  _ _ (proj2 (_.(md_invariant))))) as ( one & hone ).
        elim (project_inj hone project_one NEQ).
        apply project_in in hone. exact (Interval.const_exact _ _ hone).
      rewrite H200.
      rewrite left_identity. intro H. rewrite H in Hp2.
      assert (Hp1' := project_in Hp1). assert (Hp2' := project_in Hp2).
      destruct (in_bit Hp1') as [H11|H11]; destruct (in_bit Hp2') as [H12|H12];
        subst p1; subst p2; clear Hp1'; clear Hp2';
        [exists 0%Z|exists 0%Z|exists 0%Z|exists 1%Z]; split; auto;
        rewrite (LC.den_muln F);
        change bit with (mul bit bit); rewrite (project_mul Hp2 Hp1); auto;
        reflexivity.
    * intros H1 INV ω. hsplit. repeat subst. simpl in *. split.
      + intros β hβ v. simpl.
        destruct (INV _ _ (hm _ _ hβ) v) as (H10 & H11 & H12). auto.
      + intros β hβ p1 p2 Hp1 Hp2.
        assert (Hp1' := project_in Hp1). assert (Hp2' := project_in Hp2).
        assert (H13: ((next_root σ) < Pos.succ (next_root σ))%positive) by (clear; Psatz.lia).
        generalize (π₃ hβ _ H13). simpl. rewrite !M.gss. intro H14.
        specialize (H14 _ _ _ (refl_equal _) (refl_equal _) (refl_equal _)).
        destruct (in_bit Hp1') as [H11|H11]; destruct (in_bit Hp2') as [H12|H12];
        subst p1; subst p2; clear Hp1'; clear Hp2';
        [exists 0%Z|exists 0%Z|exists 0%Z|exists 1%Z]; split; auto;
        rewrite <- H14; change bit with (mul bit bit);
        rewrite (project_mul Hp1 Hp2); auto; reflexivity.
Qed.

Lemma bitwise_first_nil_and : ∀ bs bs' σ σ' Hincr,
  bitwise_first_nil bitAnd bs σ = STm.Result bs' σ' Hincr →
  (∀ ω, invariant ω σ) →
  (length bs ≤ Int.wordsize) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
      (length bs' ≤ Int.wordsize) ∧
      project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some 0%Z.
Proof.
  induction bs as [ |lc2 bs IHbs]; simpl; intros bs' σ σ' Hicr H INV H20 ω.
  - hsplit. repeat subst. simpl in *.
    split; auto.
    intros β hβ _. split; try constructor.
    * Psatz.lia.
    * rewrite (LC.den_zero F). rewrite <- f_of_z_zero. apply project_const.
  - hsplit. repeat subst. simpl in *.
    assert (H10:=bitAnd_spec _ _ H0 INV).
    assert (H21: length bs ≤ Int.wordsize) by (clear -H20; Psatz.lia).
    specialize (IHbs _ _ _ _ H2 (λ ω, proj1 (H10 ω)) H21 ω).
    destruct IHbs as (IH1 & IH2).
    split; auto.
    intros β hβ H. inversion H. subst. clear H.
    pose proof (solution_monotone Hincr'0) as hm'0.
    destruct (H10 ω) as (_ & H11).
    destruct (not_none H4) as (p & Hp).
    specialize (H11 _ (hm'0 _ _ hβ) 0%Z p).
    rewrite (LC.den_zero F), project_zero in H11.
    specialize (H11 (refl_equal _) Hp). destruct H11 as (p0 & Hp0 & Hp0').
    destruct (IH2 _ hβ H5).
    apply bitwise_first_nil_length in H2.
    split.
    * constructor; auto.
    rewrite Hp0; discriminate.
    * rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
      rewrite (project_add (y':=0) Hp0).
      + revert Hp0'. destruct p as [ |[p|p| ]|p]; try discriminate;
                     simpl; intro H12; inversion H12; subst p0; clear H12; auto;
                     split; auto;
                     rewrite <- H2;
                     exact H20.
      + rewrite (project_muln _ (proj2 H1)); auto.
        clear -H20 H2. unfold range_of_binary_repr. rewrite <- H2.
        apply (N.le_lt_trans _ (size (range_of_bit_numbers (S (length bs))))); auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
      + rewrite <- range_of_bit_numbers_S. rewrite  <- H2. auto using size_range_of_bit_numbers_wordsize.
Qed.

Lemma and2p1z y : Int.and (Int.repr 1) (Int.repr (2 * y)) = Int.zero.
Proof.
    apply Int.same_bits_eq.
    intros i H. rewrite Int.bits_and; auto.
    rewrite Int.bits_zero.
    rewrite ! Int.testbit_repr; auto.
    rewrite Zbits.Ztestbit_1. destruct (Coqlib.zeq _ _). 2: reflexivity.
    subst. now destruct y.
Qed.

Lemma and2p1 x y :
  Int.and (Int.repr (2 * x)) (Int.repr (1 + 2 * y)) = Int.and (Int.repr (2 * x)) (Int.repr (2 * y)).
Proof.
  rewrite <- add_repr, Int.add_is_or.
  rewrite Int.and_or_distrib.
  rewrite <- (Int.and_commut (Int.repr 1)), and2p1z.
  apply Int.or_zero_l.
  apply and2p1z.
Qed.

Lemma int_mul2 x : Int.repr (2 * x) = Int.shl (Int.repr x) Int.one.
Proof.
  clear.
  rewrite <- mul_repr.
  rewrite !(Int.mul_commut (Int.repr 2)).
  change 2 with (two_p (Int.unsigned Int.one)).
  rewrite <- !Int.shl_mul_two_p.
  reflexivity.
Qed.

Lemma bitwise_aux_and : ∀ bs1 bs2 bs σ σ' Hincr,
  bitwise_aux bitAnd bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  (length bs1 ≤ Int.wordsize) →
  (length bs2 ≤ Int.wordsize) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      (length bs ≤ Int.wordsize) ∧
      ∃ p1 p2 p,
        project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
        project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
        project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
        Int.and (Int.repr p1) (Int.repr p2) = Int.repr p.
Proof.
  induction bs1 as [ |lc1 bs1 IHbs1]; simpl; intros bs2 bs σ σ' Hincr H1 INV H2 H3 ω.
  - destruct (bitwise_first_nil_and _ H1 INV H3 ω) as (H10 & H11).
    split; auto.
    intros β hβ _ H12. specialize (H11 _ hβ H12). destruct H11 as (H11 & H300 & H13).
    split; auto.
    split; auto.
    destruct (not_none (lc_of_binary_repr_spec hβ (size_range_of_binary_repr_wordsize _ H3) H12)) as (p & Hp).
    exists 0%Z, p, 0%Z; repeat split; auto.
    rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const.
  - destruct bs2 as [ |lc2 bs2]; simpl in *.
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitAnd_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by Psatz.lia.
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H3 ω).
      generalize (bitwise_aux_length _ _ _ H5). simpl. rewrite Max.max_0_r. intro H30.
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21. inversion H20. subst. clear H20.
      specialize (IH2 _ hβ H6 H21). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      specialize (H12 _ (hm'0 _ _ hβ) _ 0%Z Hp0).
      rewrite (LC.den_zero F) in H12.
      specialize (H12 project_zero). destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + unfold range_of_binary_repr. simpl. rewrite H30.
        split; auto.
        exists (p0+2*p1)%Z, 0%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto.
          unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_muln_S.
          unfold range_of_binary_repr. simpl. auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite <- H30.
          rewrite (project_muln _ IH5); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto.
          unfold range_of_binary_repr. simpl. rewrite H30. apply size_range_of_bit_numbers_muln_S.
          unfold range_of_binary_repr. simpl. auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite Int.and_zero.
          rewrite <- add_repr. rewrite <- mul_repr.
          replace (Int.repr p3) with Int.zero. replace (Int.repr p) with Int.zero. auto.
          simpl in IH4. rewrite (LC.den_zero F), <- f_of_z_zero, project_const in IH4. inversion IH4. subst p2. clear IH4.
          rewrite Int.and_zero in IH6. exact IH6.
          destruct p0 as [ |[p0|p0| ]|p0]; try discriminate;
            simpl in Hp3'; inversion Hp3'; auto. }
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitAnd_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by (clear -H2; Psatz.lia).
      assert (H113: length bs2 ≤ Int.wordsize) by (clear -H3; Psatz.lia).
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H113 ω).
      assert (H30:=bitwise_aux_length _ _ _ H5).
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21.
      inversion H20. subst. clear H20.
      inversion H21. subst. clear H21.
      specialize (IH2 _ hβ H6 H8). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      destruct (not_none H7) as (p0' & Hp0').
      specialize (H12 _ (hm'0 _ _ hβ) _ _ Hp0 Hp0').
      destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + unfold range_of_binary_repr. simpl.
        split; [rewrite H30; apply Max.max_case_strong; auto| ].
        exists (p0+2*p1)%Z, (p0'+2*p2)%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto.
          unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_muln_S.
          unfold range_of_binary_repr. simpl. auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p2)%Z) Hp0'); auto.
          rewrite (project_muln _ IH4); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs2)%list))); auto.
          unfold range_of_binary_repr. simpl. apply size_range_of_bit_numbers_muln_S.
          unfold range_of_binary_repr. simpl. auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite (project_muln _ IH5); auto.
          unfold range_of_binary_repr. rewrite H30.
          apply Max.max_case_strong; intros _;
          [apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list)))|apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs2)%list)))]; unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite H30, <- range_of_bit_numbers_S. apply Max.max_case_strong; intros _; auto using size_range_of_bit_numbers_wordsize. }
        {
          apply project_in, in_bit in Hp0.
          apply project_in, in_bit in Hp0'.
          clear - Hp0 Hp0' Hp3' IH6.
          destruct Hp0, Hp0'; subst; simpl in Hp3'; hsplit; subst.
          rewrite ! Z.add_0_l.
          rewrite ! int_mul2, Int.and_shl, IH6. reflexivity.
          rewrite ! Z.add_0_l.
          rewrite and2p1, ! int_mul2, Int.and_shl, IH6. reflexivity.
          rewrite ! Z.add_0_l.
          rewrite Int.and_commut, and2p1, ! int_mul2, Int.and_shl, Int.and_commut, IH6. reflexivity.
          rewrite <- ! add_repr, ! int_mul2.
          rewrite !Int.add_is_or; try apply and2p1z.
          rewrite !Int.and_or_distrib.
          rewrite Int.and_commut, Int.and_or_absorb.
          f_equal.
          rewrite <- Int.add_is_or; try apply and2p1z.
          rewrite <- ! int_mul2, add_repr, Int.and_commut, and2p1.
          rewrite ! int_mul2, Int.and_shl, Int.and_commut, IH6.
          reflexivity.
        }
Qed.

Lemma bs_land_spec bs1 bs2 bs σ σ' Hincr :
  bs_land bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  (length bs1 ≤ Int.wordsize) →
  (length bs2 ≤ Int.wordsize) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      (length bs ≤ Int.wordsize) ∧
      (∃ p1 p2 p,
          project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
          project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
          project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
          Int.and (Int.repr p1) (Int.repr p2) = Int.repr p).
Proof. unfold bs_land, bitwise. intro H. eauto using bitwise_aux_and. Qed.


(** Correctness of || **)

Definition bit_Z_lor p1 p2 :=
  match p1, p2 with
  | 0%Z, 0%Z => Some 0%Z
  | 1%Z, 0%Z
  | 0%Z, 1%Z
  | 1%Z, 1%Z => Some 1%Z
  | _, _ => None
  end.

Lemma bitOr_spec b1 b2 b σ σ' Hincr :
  bitOr b1 b2 σ = STm.Result b σ' Hincr →
  (∀ ω, invariant ω σ) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      ∀ p1 p2,
      project (ω⟦ b1 ⟧ β) Interval.bit = Some p1 →
      project (ω⟦ b2 ⟧ β) Interval.bit = Some p2 →
      ∃ p, project (ω⟦ b ⟧ β) Interval.bit = Some p ∧
           bit_Z_lor p1 p2 = Some p.
Proof.
  unfold bitOr; intros H INV ω.
  pose proof (solution_monotone Hincr) as hm.
  hsplit. repeat subst. simpl in *.
  destruct (bitAnd_spec _ _ H0 INV ω) as (H1 & H2).
  split; auto.
  intros β hβ p1 p2 Hp1 Hp2.
  specialize (H2 _ hβ _ _ Hp1 Hp2).
  destruct H2 as (p & Hp & Hp'). exists (p1+(p2-p))%Z. split.
  - assert (Hp1':=Hp1). assert (Hp2':=Hp2). assert (Hp'':=Hp).
    apply project_in in Hp1. apply project_in in Hp2. apply project_in in Hp.
    apply in_bit in Hp1. apply in_bit in Hp2. apply in_bit in Hp.
    rewrite (LC.den_add F), (LC.den_sub F).
    destruct Hp1; destruct Hp2; destruct Hp; subst; try discriminate;
    repeat match goal with
    | H : project (_ ⟦ _ ⟧ _) _ = Some _ |- _ =>
      (apply project_is_zero in H || apply project_is_one in H);
      try rewrite H
    end;
    try rewrite ! left_identity;
    try rewrite ! right_identity;
    try assumption.
    assert (1 − 1 === 0) as X by ring. rewrite X, right_identity. assumption.
  - revert Hp'. clear.
    destruct p1 as [ |[p1|p1| ]|p1]; destruct p2 as [ |[p2|p2| ]|p2]; try discriminate; simpl;
      intro H; inversion H; subst p; auto.
Qed.

Lemma bitwise_first_nil_or : ∀ bs bs' σ σ' Hincr,
  bitwise_first_nil bitOr bs σ = STm.Result bs' σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
      length bs' ≤ Int.wordsize ∧
      ∃ p,
        project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
        project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p.
Proof.
  induction bs as [ |lc2 bs IHbs]; simpl; intros bs' σ σ' Hicr H INV H20 ω.
  - hsplit. repeat subst. simpl in *.
    split; auto.
    intros β hβ _. split; try constructor.
    Psatz.lia.
    exists 0%Z.
    rewrite (LC.den_zero F). rewrite <- f_of_z_zero. split; apply project_const.
  - hsplit. repeat subst. simpl in *.
    assert (H10:=bitOr_spec _ _ H0 INV).
    assert (H21: length bs ≤ Int.wordsize) by Psatz.lia.
    specialize (IHbs _ _ _ _ H2 (λ ω, proj1 (H10 ω)) H21 ω).
    destruct IHbs as (IH1 & IH2).
    split; auto.
    intros β hβ H. inversion H. subst. clear H.
    pose proof (solution_monotone Hincr'0) as hm'0.
    destruct (H10 ω) as (_ & H11).
    destruct (not_none H4) as (p & Hp).
    specialize (H11 _ (hm'0 _ _ hβ) 0%Z p).
    rewrite (LC.den_zero F), project_zero in H11.
    specialize (H11 (refl_equal _) Hp). destruct H11 as (p0 & Hp0 & Hp0').
    destruct (IH2 _ hβ H5).
    apply bitwise_first_nil_length in H2.
    split.
    * constructor; auto.
    rewrite Hp0; discriminate.
    * destruct H1 as (H1'' & p1 & H1 & H1').
      unfold range_of_binary_repr. simpl. rewrite <- H2. split; auto.
      exists (p+2*p1)%Z.
      rewrite !(LC.den_add F), !(LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite !range_of_bit_numbers_S.
      rewrite !(project_add (x':=p) (y':=(2*p1)%Z)); auto.
      + rewrite Hp0. revert Hp0'. clear.
        destruct p as [ |[p|p| ]|p]; try discriminate; auto.
      + rewrite H2. rewrite (project_muln _ H1'); auto.
        clear -lc2 H20 H2. apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs)%list))); auto; unfold range_of_binary_repr; simpl.
        rewrite H2. apply size_range_of_bit_numbers_muln_S.
        auto using size_range_of_bit_numbers_wordsize.
      + rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize.
      + rewrite (project_muln _ H1); auto.
        clear -lc2 H20 H2. apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs)%list))); auto; unfold range_of_binary_repr; simpl.
        rewrite H2. apply size_range_of_bit_numbers_muln_S.
        auto using size_range_of_bit_numbers_wordsize.
      + rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize.
Qed.

Lemma bitwise_aux_or : ∀ bs1 bs2 bs σ σ' Hincr,
  bitwise_aux bitOr bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs1 ≤ Int.wordsize →
  length bs2 ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      length bs ≤ Int.wordsize ∧
      ∃ p1 p2 p,
        project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
        project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
        project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
        Int.or (Int.repr p1) (Int.repr p2) = Int.repr p.
Proof.
  induction bs1 as [ |lc1 bs1 IHbs1]; simpl; intros bs2 bs σ σ' Hincr H1 INV H2 H3 ω.
  - destruct (bitwise_first_nil_or _ H1 INV H3 ω) as (H10 & H11).
    split; auto.
    intros β hβ _ H12. specialize (H11 _ hβ H12). destruct H11 as (H11 & H300 & p & H13 & H13').
    split; auto.
    split; auto.
    exists 0%Z, p, p; repeat split; auto.
    rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const.
    apply Int.or_zero_l.
  - destruct bs2 as [ |lc2 bs2]; simpl in *.
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitOr_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by (clear -H2; Psatz.lia).
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H3 ω).
      generalize (bitwise_aux_length _ _ _ H5). simpl. rewrite Max.max_0_r. intro H30.
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21. inversion H20. subst. clear H20.
      specialize (IH2 _ hβ H6 H21). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      specialize (H12 _ (hm'0 _ _ hβ) _ 0%Z Hp0).
      rewrite (LC.den_zero F) in H12.
      specialize (H12 project_zero). destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + split; unfold range_of_binary_repr; simpl; rewrite H30; auto.
        exists (p0+2*p1)%Z, 0%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite <- H30.
          rewrite (project_muln _ IH5); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl.
          rewrite H30. auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite Int.or_zero.
          rewrite <- !add_repr. rewrite <- !mul_repr.
          replace (Int.repr p3) with (Int.repr p0). replace (Int.repr p) with (Int.repr p1). auto.
          simpl in IH4. rewrite (LC.den_zero F), <- f_of_z_zero, project_const in IH4. inversion IH4. subst p2. clear IH4.
          rewrite Int.or_zero in IH6. exact IH6.
          destruct p0 as [ |[p0|p0| ]|p0]; try discriminate;
            simpl in Hp3'; inversion Hp3'; auto. }
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitOr_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by Psatz.lia.
      assert (H113: length bs2 ≤ Int.wordsize) by Psatz.lia.
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H113 ω).
      assert (H30:=bitwise_aux_length _ _ _ H5).
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21.
      inversion H20. subst. clear H20.
      inversion H21. subst. clear H21.
      specialize (IH2 _ hβ H6 H8). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      destruct (not_none H7) as (p0' & Hp0').
      specialize (H12 _ (hm'0 _ _ hβ) _ _ Hp0 Hp0').
      destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + split; [unfold range_of_binary_repr; simpl; rewrite H30; apply Max.max_case_strong; auto| ].
        exists (p0+2*p1)%Z, (p0'+2*p2)%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl.
          auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p2)%Z) Hp0'); auto.
          rewrite (project_muln _ IH4); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs2)%list))); auto; unfold range_of_binary_repr; simpl;
            auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite (project_muln _ IH5); auto.
          unfold range_of_binary_repr. rewrite H30.
          apply Max.max_case_strong; intros _;
            [apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list)))|apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs2)%list)))]; auto;
          unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite H30, <- range_of_bit_numbers_S. apply Max.max_case_strong; intros _; auto using size_range_of_bit_numbers_wordsize. }
        { rewrite <- !add_repr. rewrite <- !mul_repr. rewrite <- IH6.
          destruct p0 as [ |[p0|p0| ]|p0]; destruct p0' as [ |[p0'|p0'| ]|p0']; try discriminate; simpl in Hp3'; inversion Hp3'; subst; clear Hp3'.

          rewrite !Int.add_zero_l. rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          apply Int.or_shl.

          rewrite !Int.add_zero_l.
          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite Int.add_is_or; try apply and2p1z.
          rewrite Int.or_commut. rewrite Int.or_assoc.
          rewrite Int.or_shl.
          rewrite Int.add_is_or. rewrite (Int.or_commut (Int.repr p1)). reflexivity.
          unfold Int.or. rewrite <- int_mul2. apply and2p1z.

          rewrite !Int.add_zero_l.
          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite Int.add_is_or; try apply and2p1z.
          rewrite Int.or_assoc.
          rewrite Int.or_shl.
          rewrite Int.add_is_or. reflexivity.
          unfold Int.or. rewrite <- int_mul2. apply and2p1z.

          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite ! Int.add_is_or; try apply and2p1z.
          rewrite Int.or_assoc. rewrite (Int.or_commut (Int.shl _ _)).
          rewrite Int.or_assoc, <- (Int.or_assoc _ (Int.repr 1)).
          rewrite Int.or_shl, (Int.or_commut (Int.repr p2)).
          reflexivity.
        }
Qed.

Lemma bs_lor_spec bs1 bs2 bs σ σ' Hincr :
  bs_lor bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs1 ≤ Int.wordsize →
  length bs2 ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      length bs ≤ Int.wordsize ∧
      (∃ p1 p2 p,
          project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
          project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
          project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
          Int.or (Int.repr p1) (Int.repr p2) = Int.repr p).
Proof. unfold bs_lor, bitwise. intro H. eauto using bitwise_aux_or. Qed.


(** Correctness of ^^ **)

Definition bit_Z_lxor p1 p2 :=
  match p1, p2 with
  | 0%Z, 0%Z
  | 1%Z, 1%Z => Some 0%Z
  | 1%Z, 0%Z
  | 0%Z, 1%Z => Some 1%Z
  | _, _ => None
  end.

Lemma bitXor_spec b1 b2 b σ σ' Hincr :
  bitXor b1 b2 σ = STm.Result b σ' Hincr →
  (∀ ω, invariant ω σ) →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      ∀ p1 p2,
      project (ω⟦ b1 ⟧ β) Interval.bit = Some p1 →
      project (ω⟦ b2 ⟧ β) Interval.bit = Some p2 →
      ∃ p, project (ω⟦ b ⟧ β) Interval.bit = Some p ∧
           bit_Z_lxor p1 p2 = Some p.
Proof.
  unfold bitXor; intros H INV ω.
  pose proof (solution_monotone Hincr) as hm.
  hsplit. repeat subst. simpl in *.
  destruct (bitAnd_spec _ _ H0 INV ω) as (H1 & H2).
  split; auto.
  intros β hβ p1 p2 Hp1 Hp2.
  specialize (H2 _ hβ _ _ Hp1 Hp2).
  destruct H2 as (p & Hp & Hp'). exists (p1+(p2-2*p))%Z. split.
  - assert (Hp1':=Hp1). assert (Hp2':=Hp2). assert (Hp'':=Hp).
    apply project_in in Hp1. apply project_in in Hp2. apply project_in in Hp.
    apply in_bit in Hp1. apply in_bit in Hp2. apply in_bit in Hp.
    rewrite (LC.den_add F), (LC.den_sub F), (LC.den_muln F).
    destruct Hp1; destruct Hp2; destruct Hp; subst; try discriminate;
    repeat match goal with
    | H : project (_ ⟦ _ ⟧ _) _ = Some _ |- _ =>
      (apply project_is_zero in H || apply project_is_one in H);
      try rewrite H
    end;
    try rewrite ! left_identity;
    try rewrite ! right_identity.
    assert (0 − (1 + 1) × 0 === 0) as X by ring; now rewrite X.
    assert (1 − (1 + 1) × 0 === 1) as X by ring; now rewrite X.
    assert (1 + (0 − (1 + 1) × 0) === 1) as X by ring; now rewrite X.
    assert (1 +(1 − (1 + 1)) === 0) as X by ring; now rewrite X.
  - revert Hp'. clear.
    destruct p1 as [ |[p1|p1| ]|p1]; destruct p2 as [ |[p2|p2| ]|p2]; try discriminate; simpl;
      intro H; inversion H; subst p; auto.
Qed.

Lemma bitwise_first_nil_xor : ∀ bs bs' σ σ' Hincr,
  bitwise_first_nil bitXor bs σ = STm.Result bs' σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
      length bs' ≤ Int.wordsize ∧
      ∃ p,
        project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
        project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p.
Proof.
  induction bs as [ |lc2 bs IHbs]; simpl; intros bs' σ σ' Hicr H INV H20 ω.
  - hsplit. repeat subst. simpl in *.
    split; auto.
    intros β hβ _. split; try constructor.
    Psatz.lia.
    exists 0%Z.
    rewrite (LC.den_zero F). rewrite <- f_of_z_zero. split; apply project_const.
  - hsplit. repeat subst. simpl in *.
    assert (H10:=bitXor_spec _ _ H0 INV).
    assert (H21: length bs ≤ Int.wordsize) by Psatz.lia.
    specialize (IHbs _ _ _ _ H2 (λ ω, proj1 (H10 ω)) H21 ω).
    destruct IHbs as (IH1 & IH2).
    split; auto.
    intros β hβ H. inversion H. subst. clear H.
    pose proof (solution_monotone Hincr'0) as hm'0.
    destruct (H10 ω) as (_ & H11).
    destruct (not_none H4) as (p & Hp).
    specialize (H11 _ (hm'0 _ _ hβ) 0%Z p).
    rewrite (LC.den_zero F), project_zero in H11.
    specialize (H11 (refl_equal _) Hp). destruct H11 as (p0 & Hp0 & Hp0').
    destruct (IH2 _ hβ H5).
    apply bitwise_first_nil_length in H2.
    split.
    * constructor; auto.
    rewrite Hp0; discriminate.
    * destruct H1 as (H1'' & p1 & H1 & H1').
      unfold range_of_binary_repr. simpl. rewrite <- H2. split; auto.
      exists (p+2*p1)%Z.
      rewrite !(LC.den_add F), !(LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite !range_of_bit_numbers_S.
      rewrite !(project_add (x':=p) (y':=(2*p1)%Z)); auto.
      + rewrite Hp0. revert Hp0'. clear.
        destruct p as [ |[p|p| ]|p]; try discriminate; auto.
      + rewrite H2. rewrite (project_muln _ H1'); auto.
        clear -lc2 H20 H2. apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs)%list))); auto; unfold range_of_binary_repr; simpl.
        rewrite H2. auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
        auto using size_range_of_bit_numbers_wordsize.
      + rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize.
      + rewrite (project_muln _ H1); auto.
        clear -lc2 H20 H2. apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs)%list))); auto; unfold range_of_binary_repr; simpl.
        rewrite H2. auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
        auto using size_range_of_bit_numbers_wordsize.
      + rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize.
Qed.

Lemma bitwise_aux_xor : ∀ bs1 bs2 bs σ σ' Hincr,
  bitwise_aux bitXor bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs1 ≤ Int.wordsize →
  length bs2 ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      length bs ≤ Int.wordsize ∧
      ∃ p1 p2 p,
        project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
        project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
        project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
        Int.xor (Int.repr p1) (Int.repr p2) = Int.repr p.
Proof.
  induction bs1 as [ |lc1 bs1 IHbs1]; simpl; intros bs2 bs σ σ' Hincr H1 INV H2 H3 ω.
  - destruct (bitwise_first_nil_xor _ H1 INV H3 ω) as (H10 & H11).
    split; auto.
    intros β hβ _ H12. specialize (H11 _ hβ H12). destruct H11 as (H11 & H300 & p & H13 & H13').
    split; auto.
    split; auto.
    exists 0%Z, p, p; repeat split; auto.
    rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const.
    apply Int.xor_zero_l.
  - destruct bs2 as [ |lc2 bs2]; simpl in *.
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitXor_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by Psatz.lia.
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H3 ω).
      generalize (bitwise_aux_length _ _ _ H5). simpl. rewrite Max.max_0_r. intro H30.
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21. inversion H20. subst. clear H20.
      specialize (IH2 _ hβ H6 H21). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      specialize (H12 _ (hm'0 _ _ hβ) _ 0%Z Hp0).
      rewrite (LC.den_zero F) in H12.
      specialize (H12 project_zero). destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + split; unfold range_of_binary_repr; simpl; rewrite H30; auto.
        exists (p0+2*p1)%Z, 0%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl;
            auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_zero F), <- f_of_z_zero. apply project_const. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite <- H30.
          rewrite (project_muln _ IH5); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl.
          rewrite H30. auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          auto using size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite Int.xor_zero.
          rewrite <- !add_repr. rewrite <- !mul_repr.
          replace (Int.repr p3) with (Int.repr p0). replace (Int.repr p) with (Int.repr p1). auto.
          simpl in IH4. rewrite (LC.den_zero F), <- f_of_z_zero, project_const in IH4. inversion IH4. subst p2. clear IH4.
          rewrite Int.xor_zero in IH6. exact IH6.
          destruct p0 as [ |[p0|p0| ]|p0]; try discriminate;
            simpl in Hp3'; inversion Hp3'; auto. }
    * hsplit. repeat subst. simpl in *.
      assert (H10:=bitXor_spec _ _ H0 INV).
      destruct (H10 ω) as (H11 & H12).
      assert (H13: length bs1 ≤ Int.wordsize) by Psatz.lia.
      assert (H113: length bs2 ≤ Int.wordsize) by Psatz.lia.
      specialize (IHbs1 _ _ _ _ _ H5 (λ ω, proj1 (H10 ω)) H13 H113 ω).
      assert (H30:=bitwise_aux_length _ _ _ H5).
      destruct IHbs1 as (IH1 & IH2).
      split; auto.
      intros β hβ H20 H21.
      inversion H20. subst. clear H20.
      inversion H21. subst. clear H21.
      specialize (IH2 _ hβ H6 H8). destruct IH2 as (IH2 & IH2' & p1 & p2 & p & IH3 & IH4 & IH5 & IH6).
      pose proof (solution_monotone Hincr'0) as hm'0.
      destruct (not_none H4) as (p0 & Hp0).
      destruct (not_none H7) as (p0' & Hp0').
      specialize (H12 _ (hm'0 _ _ hβ) _ _ Hp0 Hp0').
      destruct H12 as (p3 & Hp3 & Hp3').
      split.
      + constructor; auto.
        rewrite Hp3; discriminate.
      + split; [unfold range_of_binary_repr; simpl; rewrite H30; apply Max.max_case_strong; auto| ].
        exists (p0+2*p1)%Z, (p0'+2*p2)%Z, (p3+2*p)%Z. repeat split.
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p1)%Z) Hp0); auto.
          rewrite (project_muln _ IH3); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list))); auto; unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p2)%Z) Hp0'); auto.
          rewrite (project_muln _ IH4); auto.
          apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs2)%list))); auto; unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize. }
        { rewrite (LC.den_add F), (LC.den_muln F). unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
          rewrite (project_add (y':=(2*p)%Z) Hp3); auto.
          rewrite (project_muln _ IH5); auto.
          unfold range_of_binary_repr. rewrite H30.
          apply Max.max_case_strong; intros _;
            [apply (N.le_lt_trans _ (size (range_of_binary_repr (lc1 :: bs1)%list)))|apply (N.le_lt_trans _ (size (range_of_binary_repr (lc2 :: bs2)%list)))]; auto;
          unfold range_of_binary_repr; simpl; auto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize.
          rewrite H30, <- range_of_bit_numbers_S. apply Max.max_case_strong; intros _; auto using size_range_of_bit_numbers_wordsize. }
        { rewrite <- !add_repr. rewrite <- !mul_repr. rewrite <- IH6.
          destruct p0 as [ |[p0|p0| ]|p0]; destruct p0' as [ |[p0'|p0'| ]|p0']; try discriminate; simpl in Hp3'; inversion Hp3'; subst; clear Hp3'.

          rewrite !Int.add_zero_l. rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          apply Int.xor_shl.

          rewrite !Int.add_zero_l.
          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite Int.add_is_xor; try apply and2p1z.
          rewrite Int.xor_commut. rewrite Int.xor_assoc.
          rewrite Int.xor_shl.
          rewrite Int.add_is_xor. rewrite (Int.xor_commut (Int.repr p1)). reflexivity.
          unfold Int.xor. rewrite <- int_mul2. apply and2p1z.

          rewrite !Int.add_zero_l.
          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite Int.add_is_xor; try apply and2p1z.
          rewrite Int.xor_assoc.
          rewrite Int.xor_shl.
          rewrite Int.add_is_xor. reflexivity.
          unfold Int.xor. rewrite <- int_mul2. apply and2p1z.

          rewrite !(Int.mul_commut (Int.repr 2)).
          change 2 with (two_p (Int.unsigned Int.one)).
          rewrite <- !Int.shl_mul_two_p.
          rewrite ! Int.add_is_xor; try apply and2p1z.
          rewrite Int.xor_assoc. rewrite (Int.xor_commut (Int.shl _ _)).
          rewrite Int.xor_assoc, <- (Int.xor_assoc _ (Int.repr 1)).
          rewrite Int.xor_shl, (Int.xor_commut (Int.repr p2)).
          reflexivity.
          apply Int.and_zero_l.
        }
Qed.

Lemma bs_lxor_spec bs1 bs2 bs σ σ' Hincr :
  bs_lxor bs1 bs2 σ = STm.Result bs σ' Hincr →
  (∀ ω, invariant ω σ) →
  length bs1 ≤ Int.wordsize →
  length bs2 ≤ Int.wordsize →
  ∀ ω, invariant ω σ' ∧ ∀ β (hβ: solution ω σ' β),
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs1 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs2 →
      List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs ∧
      length bs ≤ Int.wordsize ∧
      (∃ p1 p2 p,
          project (ω⟦ lc_of_binary_repr bs1 ⟧ β) (range_of_binary_repr bs1) = Some p1 ∧
          project (ω⟦ lc_of_binary_repr bs2 ⟧ β) (range_of_binary_repr bs2) = Some p2 ∧
          project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
          Int.xor (Int.repr p1) (Int.repr p2) = Int.repr p).
Proof. unfold bs_lxor, bitwise. intro H. eauto using bitwise_aux_xor. Qed.


(** Correctness of >> **)

Theorem shru_shru x y z:
  Int.unsigned y + Int.unsigned z < Int.modulus →
  Int.shru (Int.shru x y) z = Int.shru x (Int.add y z).
Proof.
  intro H.
  assert (H1: Int.unsigned (Int.add y z) = (Int.unsigned y + Int.unsigned z)%Z) by (unfold Int.add; apply Int.unsigned_repr; generalize (Int.unsigned_range_2 y); generalize (Int.unsigned_range_2 z); unfold Int.max_unsigned; Psatz.lia).
  apply Int.same_bits_eq; intros.
  rewrite Int.bits_shru; auto.
  destruct (Coqlib.zlt (i + Int.unsigned z) Int.zwordsize).
  - rewrite Int.bits_shru. destruct (Coqlib.zlt (i + Int.unsigned z + Int.unsigned y) Int.zwordsize).
    + rewrite Int.bits_shru; auto. rewrite Coqlib.zlt_true. f_equal. omega. omega.
    + rewrite Int.bits_shru; auto. rewrite Coqlib.zlt_false. auto. omega.
    + generalize (Int.unsigned_range_2 z); omega.
  - rewrite Int.bits_shru; auto. rewrite Coqlib.zlt_false. auto. generalize (Int.unsigned_range_2 y); omega.
Qed.


Lemma two_power_nat_compat x y :
  x ≤ y → (two_power_nat x <= two_power_nat y)%Z.
Proof.
  induction 1 as [ |y Hy IH].
  - reflexivity.
  - rewrite two_power_nat_S. generalize (Coqlib.two_power_nat_pos y).
    Psatz.lia.
Qed.

Lemma shru_lemma p b :
  0 <= p < two_power_nat (pred Int.wordsize) →
  (b = 0 ∨ b = 1)%Z →
   Int.shru (Int.add (Int.repr b) (Int.shl (Int.repr p) Int.one))
     (Int.repr 1) = Int.repr p.
Proof.
  clear.
  assert (Int.max_unsigned = 4294967295) by reflexivity.
  intros RANGE BIT.
  rewrite Int.shl_mul.
  change (Int.shl Int.one Int.one) with (Int.repr 2).
  rewrite mul_repr, add_repr.
  unfold Int.shru. f_equal.
  rewrite ! Int.unsigned_repr.
  rewrite Z.shiftr_div_pow2.
  rewrite Z.div_add.
  destruct BIT; subst; reflexivity.
  Psatz.lia.
  Psatz.lia.
  Psatz.lia.
  assert (0 <= p < 2147483648) by exact RANGE.
  Psatz.lia.
Qed.

Lemma bs_shru_aux_spec ω σ β (Hβ : solution ω σ β) :
  ∀ n bs, Z.of_nat n < Int.modulus →
  let bs' := bs_shru_aux bs n in
  length bs ≤ Int.wordsize →
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
  length bs' ≤ Int.wordsize ∧
  (∃ p p',
      project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
      project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p' ∧
      Int.shru (Int.repr p) (Int.repr (Z.of_nat n)) = Int.repr p').
Proof.
  induction n as [ |n IHn]; simpl.
  - intros bs _ Hsize Hbs.
    repeat split; auto.
    destruct (not_none (lc_of_binary_repr_spec Hβ (size_range_of_binary_repr_wordsize _ Hsize) Hbs)) as [p Hp].
    exists p, p; repeat split; auto.
    apply Int.shru_zero.
  - intros [ |b bs]; simpl; intros Hmod Hsize Hbs.
    * repeat split; auto.
      exists 0%Z, 0%Z. rewrite (LC.den_zero F). rewrite <- f_of_z_zero. rewrite project_const. repeat split; auto.
      unfold Int.shru. change (Int.unsigned (Int.repr 0)) with 0%Z. rewrite Z.shiftr_0_l. reflexivity.
    * inversion Hbs. subst x l. clear Hbs.
      assert (Hmod': Z.of_nat n < Int.modulus) by (clear -Hmod; Psatz.lia).
      assert (Hsize': length bs ≤ Int.wordsize) by (clear -Hsize; Psatz.lia).
      assert (Hsize'':(size (muln 2 (range_of_binary_repr bs)) < N.pos large_prime)%N) by (unfold range_of_binary_repr in *; apply (N.le_lt_trans _ (size (range_of_bit_numbers (S (length bs))))); eauto using size_range_of_bit_numbers_muln_S, size_range_of_bit_numbers_wordsize).
      destruct (IHn _ Hmod' Hsize' H2) as (H10 & H11 & p & p' & H12 & H13 & H14).
      repeat split; auto.
      destruct (not_none H1) as [b0 Hb0].
      exists (b0 + 2*p)%Z, p'. split.
      + rewrite (LC.den_add F), (LC.den_muln F).
        unfold range_of_binary_repr. simpl. rewrite range_of_bit_numbers_S.
        rewrite (project_add (y':=2*p) Hb0); auto.
        rewrite <- range_of_bit_numbers_S. auto using size_range_of_bit_numbers_wordsize.
      + split; auto. Opaque Z.mul.
        rewrite Zpos_P_of_succ_nat, <- Z.add_1_r.
        rewrite <- H14. rewrite (Z.add_comm _ 1%Z). rewrite <- (add_repr 1%Z).
        rewrite <- shru_shru; [ |change (Int.unsigned (Int.repr 1)) with 1%Z; rewrite Int.unsigned_repr_eq; rewrite Zmod_small; auto using Zle_0_nat; Psatz.lia].
        f_equal.

        rewrite <- add_repr, <- mul_repr.
        rewrite !(Int.mul_commut (Int.repr 2)).
        change 2 with (two_p (Int.unsigned Int.one)).
        rewrite <- !Int.shl_mul_two_p.

        assert (H20: 0 <= p < two_power_nat (pred Int.wordsize)).
          clear -project_in H12 Hsize.
          assert (H: (length bs < Int.wordsize)%nat) by Psatz.lia.
          apply project_in in H12. unfold denote, is_true in H12. rewrite andb_true_iff in H12. rewrite ! Z.leb_le in H12. unfold range_of_binary_repr, range_of_bit_numbers in H12. simpl in H12. unfold sup_of_bit_numbers in H12.
          apply Nat.lt_le_pred, two_power_nat_compat in H. Psatz.lia.

        apply project_in, in_bit in Hb0.

        clear -H20 Hb0.
        now apply shru_lemma.
Qed.


Lemma bs_shru_spec ω σ β (Hβ : solution ω σ β) bs i :
  let bs' := bs_shru bs i in
  length bs ≤ Int.wordsize →
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs →
  List.Forall (fun b => project (ω⟦ b ⟧ β) Interval.bit ≠ None) bs' ∧
  length bs' ≤ Int.wordsize ∧
  (∃ p p',
      project (ω⟦ lc_of_binary_repr bs ⟧ β) (range_of_binary_repr bs) = Some p ∧
      project (ω⟦ lc_of_binary_repr bs' ⟧ β) (range_of_binary_repr bs') = Some p' ∧
      Int.shru (Int.repr p) i = Int.repr p').
Proof.
  intros bs' H1 H2. subst bs'.
  assert (H3: Z.of_nat (Z.to_nat (Int.unsigned i)) < Int.modulus).
    destruct (Int.unsigned_range i) as [H3 H4].
    now rewrite Z2Nat.id.
  destruct (bs_shru_aux_spec Hβ (Z.to_nat (Int.unsigned i)) H3 H1 H2) as (H10 & H11 & p & p' & H12 & H13 & H14).
  repeat split; auto.
  exists p, p'; repeat split; auto.
  rewrite <- H14.
  destruct (Int.unsigned_range i) as [H20 H21].
  rewrite Z2Nat.id; auto.
  now rewrite Int.repr_unsigned.
Qed.


(* Correctness of all the arithmetic operations *)

Lemma of_int_sound :
   ∀ (i : int) (σ : cqap f) (x : positive) (σ' : cqap f)
   (Hincr : incr σ σ'),
   Arith.of_int (cQAP f_of_z) i σ = STm.Result x σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.repr (ρ x) = i ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    simpl. unfold fresh_lc.
    intros i σ x σ' Hincr H ρ (β & ω & hβ &hv) INV.
    flatten
    hsplit. subst. simpl in *. split.
    - now rewrite <- (hv _ _ (get_env_gss _ _ _) _ (lc_of_f_denote (Int.unsigned i) hβ)), Int.repr_unsigned.
    - clear ω β hβ hv.
      refine (fresh_lc_invariant _ _ _ INV).
      * intros ω β hβ. simpl. rewrite (lc_of_f_denote _ hβ). intro. hsplit.
      * discriminate.
      * discriminate.
Qed.

Lemma to_int_sound :
   ∀ (x : positive) (σ : cqap f) (i : int) (σ' : cqap f)
   (Hincr : incr σ σ'),
   Arith.to_int (cQAP f_of_z) x σ = STm.Result i σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.repr (ρ x) = i ∧ (∀ ω : LC.var, invariant ω σ').
  Proof.
    unfold Arith.to_int. simpl.
    unfold get_range.
    intros x σ i σ' Hincr H ρ (β & ω & Hsol & H0) H1.
    hsplit. subst. subst.
    generalize (is_const_elim x0). destruct (is_const _). 2: hsplit.
    intros X. specialize (X _ eq_refl). destruct X as [ X Y ].
    flatten_eq hsplit; subst; refine (conj _ H1);
    specialize (H1 _ _ Hsol x);
    assert ( GE := (get_env_Some_simpl _ _ Heqo0)) ;
    destruct H1 as [ U _ ];
    specialize (U _ GE);
    hsplit; simpl in *; subst;
    (rewrite <- (H0 x _ GE z);
      [ reflexivity | simpl ]);
    rewrite U; f_equal; apply Y, (project_in U).
Qed.


Lemma add_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) (σ' : cqap f)
   (Hincr : incr σ σ'),
   Arith.add (cQAP f_of_z) x y σ = STm.Result z σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.add (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
       ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -project_add project_proper F.
    simpl. unfold fresh_lc, QAP.get_lc.
    intros x y σ z σ' Hincr H ρ (β & ω & hβ & hv) INV.
    hsplit. repeat subst. simpl in *. destruct (sumbool_of_bool _); hsplit. flatten_eq hsplit. subst. simpl in *.
    split; [
        generalize (hv _ _ (get_env_gss _ _ _))
      | clear ω β hβ hv;
        refine (fresh_lc_invariant _ _ _ INV);
        try discriminate;
        intros β ω hβ
 ]; simpl;
    repeat
    match goal with
    | H : _ ! ?x = Some ?v |- _ =>
      let q := fresh "q" in
      let hq := fresh "hq" in
      apply get_env_Some_simpl in H;
      destruct (proj1 (INV _ _ hβ _) _ H) as (q & hq & _);
      try (
        generalize (hv x (lc_of_value v) (eq_trans (get_env_gso _ _ (lt_ne (get_env_inv _ _ H))) H) q hq);
        intros ->);
      clear H
    end;
    repeat
      match goal with
      | H : lc_of_value _ = (_ , _) |- _ =>
        rewrite H in *; simpl in *; clear H
      end;
    match goal with
    | H : (size (add ?a ?b) <? _)%N = true, Pa : project _ ?a = _, Pb : project _ ?b = _  |- _ =>
      pose proof (project_add Pa Pb (proj1 (N.ltb_lt _ _) H)) as X
    end.

    setoid_rewrite (LC.den_add F). setoid_rewrite X.
    intros H1. specialize (H1 _ eq_refl). rewrite add_repr. congruence.
    setoid_rewrite (LC.den_add F). simpl. congruence.
Qed.

Lemma sub_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) (σ' : cqap f)
   (Hincr : incr σ σ'),
   Arith.sub (cQAP f_of_z) x y σ = STm.Result z σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.sub (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
       ∧ (∀ ω : LC.var, invariant ω σ').
 Proof.
    clear -project_sub F project_proper.
    simpl. unfold fresh_lc, QAP.get_lc.
    intros x y σ z σ' Hincr H ρ (β & ω & hβ & hv) INV.
    hsplit. repeat subst. simpl in *. destruct (sumbool_of_bool _); hsplit. flatten_eq hsplit. subst. simpl in *.
    split; [
        generalize (hv _ _ (get_env_gss _ _ _))
      | clear β ω hβ hv;
        refine (fresh_lc_invariant _ _ _ INV);
        try discriminate;
        intros β ω hβ
 ];
    repeat
    match goal with
    | H : _ ! ?x = Some ?v |- _ =>
      let q := fresh "q" in
      let hq := fresh "hq" in
      apply get_env_Some_simpl in H;
      destruct (proj1 (INV _ _ hβ _) _ H) as (q & hq & _);
      try (
        generalize (hv x (lc_of_value v) (eq_trans (get_env_gso _ _ (lt_ne (get_env_inv _ _ H))) H) q hq);
        intros ->);
      clear H
    end;
    repeat
      match goal with
      | H : lc_of_value _ = (_ , _) |- _ =>
        rewrite H in *; simpl in *; clear H
      end;
    match goal with
    | H : (size (Interval.sub ?a ?b) <? _)%N = true, Pa : project _ ?a = _, Pb : project _ ?b = _  |- _ =>
      pose proof (project_sub Pa Pb (proj1 (N.ltb_lt _ _) H)) as X
    end.

    setoid_rewrite (LC.den_sub F). setoid_rewrite X.
    intros H1. specialize (H1 _ eq_refl). rewrite sub_repr. congruence.
    setoid_rewrite (LC.den_sub F). simpl. congruence.
Qed.

Lemma mul_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) (b : bool)
   (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.mul (cQAP f_of_z) x y σ = STm.Result (z, b) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.mul (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -project_mul project_muln F project_proper project_in π₃.
    simpl. unfold fresh_lc, QAP.get_lc, fresh_eqn, fresh_wire.
    intros x y σ z b σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    repeat (hsplit; subst). simpl in *.
    repeat
    match goal with
    | H : match is_const ?x with _ => _ end _ = _ |- _ =>
      let CST := fresh "hcst" in
      (destruct (is_const x) as [ ? | ] eqn: CST; [ apply is_const_elim in CST | ])
    end; destruct (sumbool_of_bool _); hsplit; flatten_eq hsplit; repeat subst; simpl in *;
    (split; [
       generalize (hv _ _ (get_env_gss _ _ _))
       | clear ω β hβ hv;
         (refine (fresh_lc_invariant _ _ _ INV); try discriminate; intros ω β hβ)
           ||
         (refine (fresh_lc_invariant _ _ _ _); try discriminate; intros ω β hβ)
         ]);
    repeat match goal with
     | H : _ ! ?x = Some ?v |- _ =>
      let q := fresh "q" in
      let hq := fresh "hq" in
      apply get_env_Some_simpl in H;
      destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ H) as (q & hq & _);
      try (pose proof (hv x (lc_of_value v) (eq_trans (get_env_gso _ _ (lt_ne (get_env_inv _ _ H))) H) q hq)); clear H
    | H : ∀ _, _ → _ = _, K : project _ ?x = Some _ |- _ => pose proof (H _ (project_in K)); clear H
  | H : lc_of_value _ = (_ , _) |- _ =>
    rewrite H in *; simpl in *; clear H
    end; subst; eauto;
    match goal with
    | H : (size (muln ?a ?b) <? _)%N = true, Pb : project _ ?b = _ |- _ =>
      pose proof (project_muln a Pb (proj1 (N.ltb_lt _ _) H)) as X
    | H : (size (mul ?a ?b) <? _)%N = true, Pa : project _ ?a = _, Pb : project _ ?b = _ |- _ =>
      pose proof (project_mul Pa Pb (proj1 (N.ltb_lt _ _) H)) as X
    end.
    subst. intros Q.
    setoid_rewrite (LC.den_muln F) in Q. setoid_rewrite X in Q. specialize (Q _ eq_refl).
    rewrite mul_repr. congruence.
    simpl. rewrite (LC.den_muln F).
    congruence.
    subst. intros Q.
    setoid_rewrite (LC.den_muln F) in Q. setoid_rewrite X in Q. specialize (Q _ eq_refl).
    rewrite Int.mul_commut, mul_repr. congruence.
    simpl. rewrite (LC.den_muln F).
    congruence.

    unfold lc_of_wire.
    setoid_rewrite <- (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
    intros Q. rewrite X in Q. specialize (Q _ eq_refl). rewrite mul_repr. congruence.
    setoid_rewrite <- (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
    simpl. congruence.

    eapply INV. eauto.
Qed.


Lemma land_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) h
   (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.land (cQAP f_of_z) x y σ = STm.Result (z, h) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.and (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    simpl. unfold fresh_lc.
    intros x y σ z h σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. repeat subst. simpl in *.
    destruct (sumbool_of_bool _); hsplit.
    repeat subst. simpl in *.
    pose proof (solution_monotone Hincr'0) as hm'0.
    pose proof (solution_monotone Hincr'1) as hm'1.
    assert (H10:=get_binary_decomposition_spec _ H0 INV).
    assert (H11:=get_binary_decomposition_spec _ H2 (λ ω, proj1 (H10 ω))).
    destruct (H10 ω) as (_ & H30).
    specialize (H30 _ (hm'0 _ _ (hm'1 _ _ hβ))).
    destruct H30 as (H20 & H30 & lc & Hincr'2 & H31 & p & p' & H32 & H33 & H34).
    destruct (H11 ω) as (_ & H40).
    specialize (H40 _ (hm'1 _ _ hβ)).
    destruct H40 as (H21 & H40 & lc0 & Hincr'3 & H41 & p0 & p'0 & H42 & H43 & H44).
    destruct (bs_land_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω) as (H12 & H13).
    specialize (H13 _ hβ H30 H40).
    destruct H13 as (H13 & p1 & p2 & p'' & p17 & H14 & H15 & H16 & H17).
    unfold QAP.get_lc in H31. unfold QAP.get_lc in H41.
    hsplit. repeat subst. simpl in *.
    case_eq ((env s''3) ! x); [intro vx| ]; intro Hvx; rewrite Hvx in H31; hsplit; repeat subst; simpl in *.
    case_eq ((env s''2) ! y); [intro vy| ]; intro Hvy; rewrite Hvy in H41; hsplit; repeat subst; simpl in *.
    pose proof (solution_monotone Hincr') as hm'.
    generalize (hv x (lc_of_value vx)). unfold get_env.
    assert (x < next_var s''3)%positive.
    assert (H200: (x < next_var s''3)%positive ∨ (next_var s''3 <= x)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''3)) in Hvx; auto; discriminate.
    assert (x < next_var s''2)%positive.
    apply (Pos.lt_le_trans _ (next_var s''3)); auto.
    exact (proj1 (proj2 (proj2 Hincr'))).
    assert (x < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (x < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (x ≠ next_var s''1) by (clear -H5; Psatz.lia).
    assert (y < next_var s''2)%positive.
    assert (H200: (y < next_var s''2)%positive ∨ (next_var s''2 <= y)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''2)) in Hvy; auto; discriminate.
    assert (y < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (y < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (y ≠ next_var s''1) by (clear -H9; Psatz.lia).
    rewrite (M.gso _ _ H6).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'))))))); auto.
    rewrite Hvx. intro H50. specialize (H50 refl_equal p).
    rewrite H31 in H50. simpl in H50. specialize (H50 H32). subst p.
    generalize (hv y (lc_of_value vy)). unfold get_env.
    rewrite (M.gso _ _ H18).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite Hvy. intro H50. specialize (H50 refl_equal p0).
    rewrite H41 in H50. simpl in H50. specialize (H50 H42). subst p0.
    rewrite H34, H44.
    generalize (hv (next_var s''1) (lc_of_binary_repr x4, range_of_binary_repr x4)). unfold get_env.
    rewrite M.gss. simpl. intro H50. specialize (H50 refl_equal _ H16). subst p17.
    rewrite H14 in H33. inversion H33. subst.
    rewrite H15 in H43. inversion H43. subst.
    split; auto.
    refine (fresh_lc_invariant _ _ _ _); simpl.
    intros ω0 β0 hβ0. clear H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & pp1 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_land_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & p114 & H114 & H115 & H116 & H117).
    rewrite H116. discriminate.
    intros ω0 β0 bs hβ0 Hbs. inversion Hbs. subst bs. clear Hbs H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_land_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & H114 & H115 & H116 & H117).
    exact p11.
    intros ω0 β0 bs H300 hβ0. clear H20 H21.
    inversion H300. subst bs. clear H300.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_land_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140). exact (proj1 H113).
    intros ω0 β0 hβ0.
    destruct (bs_land_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    auto.
Qed.


Lemma lor_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) h
   (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.lor (cQAP f_of_z) x y σ = STm.Result (z, h) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.or (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    simpl. unfold fresh_lc.
    intros x y σ z h σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. repeat subst. simpl in *.
    destruct (sumbool_of_bool _); hsplit.
    repeat subst. simpl in *.
    pose proof (solution_monotone Hincr'0) as hm'0.
    pose proof (solution_monotone Hincr'1) as hm'1.
    assert (H10:=get_binary_decomposition_spec _ H0 INV).
    assert (H11:=get_binary_decomposition_spec _ H2 (λ ω, proj1 (H10 ω))).
    destruct (H10 ω) as (_ & H30).
    specialize (H30 _ (hm'0 _ _ (hm'1 _ _ hβ))).
    destruct H30 as (H20 & H30 & lc & Hincr'2 & H31 & p & p' & H32 & H33 & H34).
    destruct (H11 ω) as (_ & H40).
    specialize (H40 _ (hm'1 _ _ hβ)).
    destruct H40 as (H21 & H40 & lc0 & Hincr'3 & H41 & p0 & p'0 & H42 & H43 & H44).
    destruct (bs_lor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω) as (H12 & H13).
    specialize (H13 _ hβ H30 H40).
    destruct H13 as (H13 & p1 & p2 & p'' & p17 & H14 & H15 & H16 & H17).
    unfold QAP.get_lc in H31. unfold QAP.get_lc in H41.
    hsplit. repeat subst. simpl in *.
    case_eq ((env s''3) ! x); [intro vx| ]; intro Hvx; rewrite Hvx in H31; hsplit; repeat subst; simpl in *.
    case_eq ((env s''2) ! y); [intro vy| ]; intro Hvy; rewrite Hvy in H41; hsplit; repeat subst; simpl in *.
    pose proof (solution_monotone Hincr') as hm'.
    generalize (hv x (lc_of_value vx)). unfold get_env.
    assert (x < next_var s''3)%positive.
    assert (H200: (x < next_var s''3)%positive ∨ (next_var s''3 <= x)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''3)) in Hvx; auto; discriminate.
    assert (x < next_var s''2)%positive.
    apply (Pos.lt_le_trans _ (next_var s''3)); auto.
    exact (proj1 (proj2 (proj2 Hincr'))).
    assert (x < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (x < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (x ≠ next_var s''1) by (clear -H5; Psatz.lia).
    assert (y < next_var s''2)%positive.
    assert (H200: (y < next_var s''2)%positive ∨ (next_var s''2 <= y)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''2)) in Hvy; auto; discriminate.
    assert (y < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (y < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (y ≠ next_var s''1) by (clear -H9; Psatz.lia).
    rewrite (M.gso _ _ H6).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'))))))); auto.
    rewrite Hvx. intro H50. specialize (H50 refl_equal p).
    rewrite H31 in H50. simpl in H50. specialize (H50 H32). subst p.
    generalize (hv y (lc_of_value vy)). unfold get_env.
    rewrite (M.gso _ _ H18).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite Hvy. intro H50. specialize (H50 refl_equal p0).
    rewrite H41 in H50. simpl in H50. specialize (H50 H42). subst p0.
    rewrite H34, H44.
    generalize (hv (next_var s''1) (lc_of_binary_repr x4, range_of_binary_repr x4)). unfold get_env.
    rewrite M.gss. simpl. intro H50. specialize (H50 refl_equal _ H16). subst p17.
    rewrite H14 in H33. inversion H33. subst.
    rewrite H15 in H43. inversion H43. subst.
    split; auto.
    refine (fresh_lc_invariant _ _ _ _); simpl.
    intros ω0 β0 hβ0. clear H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & pp1 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & p114 & H114 & H115 & H116 & H117).
    rewrite H116. discriminate.
    intros ω0 β0 bs hβ0 Hbs. inversion Hbs. subst bs. clear Hbs H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & H114 & H115 & H116 & H117).
    exact p11.
    intros ω0 β0 bs H300 hβ0. clear H20 H21.
    inversion H300. subst bs. clear H300.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140). exact (proj1 H113).
    intros ω0 β0 hβ0.
    destruct (bs_lor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    auto.
Qed.


Lemma lxor_sound :
   ∀ (x y : positive) (σ : cqap f) (z : positive) h
   (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.lxor (cQAP f_of_z) x y σ = STm.Result (z, h) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.xor (Int.repr (ρ x)) (Int.repr (ρ y)) = Int.repr (ρ z)
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    simpl. unfold fresh_lc.
    intros x y σ z h σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. repeat subst. simpl in *.
    destruct (sumbool_of_bool _); hsplit.
    repeat subst. simpl in *.
    pose proof (solution_monotone Hincr'0) as hm'0.
    pose proof (solution_monotone Hincr'1) as hm'1.
    assert (H10:=get_binary_decomposition_spec _ H0 INV).
    assert (H11:=get_binary_decomposition_spec _ H2 (λ ω, proj1 (H10 ω))).
    destruct (H10 ω) as (_ & H30).
    specialize (H30 _ (hm'0 _ _ (hm'1 _ _ hβ))).
    destruct H30 as (H20 & H30 & lc & Hincr'2 & H31 & p & p' & H32 & H33 & H34).
    destruct (H11 ω) as (_ & H40).
    specialize (H40 _ (hm'1 _ _ hβ)).
    destruct H40 as (H21 & H40 & lc0 & Hincr'3 & H41 & p0 & p'0 & H42 & H43 & H44).
    destruct (bs_lxor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω) as (H12 & H13).
    specialize (H13 _ hβ H30 H40).
    destruct H13 as (H13 & p1 & p2 & p'' & p17 & H14 & H15 & H16 & H17).
    unfold QAP.get_lc in H31. unfold QAP.get_lc in H41.
    hsplit. repeat subst. simpl in *.
    case_eq ((env s''3) ! x); [intro vx| ]; intro Hvx; rewrite Hvx in H31; hsplit; repeat subst; simpl in *.
    case_eq ((env s''2) ! y); [intro vy| ]; intro Hvy; rewrite Hvy in H41; hsplit; repeat subst; simpl in *.
    pose proof (solution_monotone Hincr') as hm'.
    generalize (hv x (lc_of_value vx)). unfold get_env.
    assert (x < next_var s''3)%positive.
    assert (H200: (x < next_var s''3)%positive ∨ (next_var s''3 <= x)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''3)) in Hvx; auto; discriminate.
    assert (x < next_var s''2)%positive.
    apply (Pos.lt_le_trans _ (next_var s''3)); auto.
    exact (proj1 (proj2 (proj2 Hincr'))).
    assert (x < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (x < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (x ≠ next_var s''1) by (clear -H5; Psatz.lia).
    assert (y < next_var s''2)%positive.
    assert (H200: (y < next_var s''2)%positive ∨ (next_var s''2 <= y)%positive) by (clear; Psatz.lia).
    destruct H200 as [H200|H200]; auto.
    rewrite (proj2 (wf s''2)) in Hvy; auto; discriminate.
    assert (y < next_var s''0)%positive.
    apply (Pos.lt_le_trans _ (next_var s''2)); auto.
    exact (proj1 (proj2 (proj2 Hincr'0))).
    assert (y < next_var s''1)%positive.
    apply (Pos.lt_le_trans _ (next_var s''0)); auto.
    exact (proj1 (proj2 (proj2 Hincr'1))).
    assert (y ≠ next_var s''1) by (clear -H9; Psatz.lia).
    rewrite (M.gso _ _ H6).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'))))))); auto.
    rewrite Hvx. intro H50. specialize (H50 refl_equal p).
    rewrite H31 in H50. simpl in H50. specialize (H50 H32). subst p.
    generalize (hv y (lc_of_value vy)). unfold get_env.
    rewrite (M.gso _ _ H18).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'1))))))); auto.
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'0))))))); auto.
    rewrite Hvy. intro H50. specialize (H50 refl_equal p0).
    rewrite H41 in H50. simpl in H50. specialize (H50 H42). subst p0.
    rewrite H34, H44.
    generalize (hv (next_var s''1) (lc_of_binary_repr x4, range_of_binary_repr x4)). unfold get_env.
    rewrite M.gss. simpl. intro H50. specialize (H50 refl_equal _ H16). subst p17.
    rewrite H14 in H33. inversion H33. subst.
    rewrite H15 in H43. inversion H43. subst.
    split; auto.
    refine (fresh_lc_invariant _ _ _ _); simpl.
    intros ω0 β0 hβ0. clear H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & pp1 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lxor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & p114 & H114 & H115 & H116 & H117).
    rewrite H116. discriminate.
    intros ω0 β0 bs hβ0 Hbs. inversion Hbs. subst bs. clear Hbs H20 H21.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lxor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140).
    destruct H113 as (H113 & p11 & p21 & p''1 & H114 & H115 & H116 & H117).
    exact p11.
    intros ω0 β0 bs H300 hβ0. clear H20 H21.
    inversion H300. subst bs. clear H300.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ (hm'1 _ _ hβ0))).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (H11 ω0) as (_ & H140).
    specialize (H140 _ (hm'1 _ _ hβ0)).
    destruct H140 as (H21 & H140 & lc0100 & Hincr'31 & H141 & p01 & p'01 & H142 & H143 & H144).
    destruct (bs_lxor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    specialize (H113 _ hβ0 H130 H140). exact (proj1 H113).
    intros ω0 β0 hβ0.
    destruct (bs_lxor_spec _ _ H4 (λ ω, proj1 (H11 ω)) H20 H21 ω0) as (H112 & H113).
    auto.
Qed.


Lemma shru_sound :
   ∀ (x : positive) (i : int) (σ : cqap f) (z : positive)
   (h : option nat) (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.shru (cQAP f_of_z) x i σ = STm.Result (z, h) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.shru (Int.repr (ρ x)) i = Int.repr (ρ z)
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    simpl. unfold fresh_lc.
    intros x i σ z h σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. repeat subst. simpl in *.
    destruct (sumbool_of_bool _); hsplit.
    pose proof (solution_monotone Hincr'0) as hm'0.
    repeat subst. simpl in *.
    assert (H10:=get_binary_decomposition_spec _ H0 INV).
    destruct (H10 ω) as (_ & H30).
    specialize (H30 _ (hm'0 _ _ hβ)).
    destruct H30 as (H20 & H30 & lc & Hincr'2 & H31 & p & p' & H32 & H33 & H34).
    destruct (bs_shru_spec hβ i H20 H30) as (H40 & H41 & q & q' & H42 & H43 & H44).
    unfold QAP.get_lc in H31.
    hsplit. repeat subst. simpl in *.
    case_eq ((env s''0) ! x); [intro vx| ]; intro Hvx; rewrite Hvx in H31; hsplit; repeat subst; simpl in *.
    pose proof (solution_monotone Hincr') as hm'.
    generalize (hv x (lc_of_value vx)). unfold get_env.
    assert (x < next_var s''0)%positive.
      assert (H200: (x < next_var s''0)%positive ∨ (next_var s''0 <= x)%positive) by (clear; Psatz.lia).
      destruct H200 as [H200|H200]; auto.
      rewrite (proj2 (wf s''0)) in Hvx; auto; discriminate.
    assert (x < next_var s'')%positive.
      apply (Pos.lt_le_trans _ (next_var s''0)); auto.
      exact (proj1 (proj2 (proj2 Hincr'))).
    assert (x ≠ next_var s'') by (clear -H1; Psatz.lia).
    rewrite (M.gso _ _ H2).
    rewrite <- (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 (proj2 Hincr'))))))); auto.
    rewrite Hvx. intro H50. specialize (H50 refl_equal p).
    rewrite H31 in H50. simpl in H50. specialize (H50 H32). subst p.
    rewrite H34.
    generalize (hv (next_var s'') (lc_of_binary_repr (bs_shru x0 i), range_of_binary_repr (bs_shru x0 i))). unfold get_env.
    rewrite M.gss. simpl. intro H50. specialize (H50 refl_equal _ H43). subst q'.
    rewrite H42 in H33. inversion H33. subst p'.
    split; auto.
    refine (fresh_lc_invariant _ _ _ _); simpl.
    intros ω0 β0 hβ0. clear H20.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ hβ0)).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & pp1 & p'1 & H132 & H133 & H134).
    destruct (bs_shru_spec hβ0 i H20 H130) as (H140 & H141 & q1 & q'1 & H142 & H143 & H144).
    rewrite H143. discriminate.
    intros ω0 β0 bs hβ0 Hbs. inversion Hbs. subst bs. clear Hbs H20.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ hβ0)).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (bs_shru_spec hβ0 i H20 H130) as (H140 & H141 & q1 & q'1 & H142 & H143 & H144).
    exact H141.
    intros ω0 β0 bs H300 hβ0. clear H20.
    inversion H300. subst bs. clear H300.
    destruct (H10 ω0) as (_ & H130).
    specialize (H130 _ (hm'0 _ _ hβ0)).
    destruct H130 as (H20 & H130 & lc100 & Hincr'21 & H131 & p17 & p'1 & H132 & H133 & H134).
    destruct (bs_shru_spec hβ0 i H20 H130) as (H140 & H141 & q1 & q'1 & H142 & H143 & H144).
    auto.
    intros ω0 β0 hβ0.
    destruct (H10 ω0) as (H130 & _).
    auto.
Qed.

Ltac subst_simpl :=
  repeat match goal with H : ?a = _ |- _ => subst a; simpl in * end;
  repeat match goal with H : _ = ?b |- _ => subst b; simpl in * end.

Lemma cmp_sound :
   ∀ (s : bool) (c : comparison) (x y : positive) (σ : cqap f)
   (z : positive) (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.cmp (cQAP f_of_z) s c x y σ = STm.Result z σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → (if (if s then Int.cmp else Int.cmpu) c (Int.repr (ρ x))
               (Int.repr (ρ y))
          then Int.one
          else Int.zero) = Int.repr (ρ z) ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -F π₃ project_one' project_zero project_in project_inj project_one project_fun project_proper project_const.
    simpl. unfold fresh_lc, fresh_eqn, fresh_wire, QAP.get_lc, lc_one.
    intros s c a b σ z σ' Hincr H ρ (β & ω & hβ & hv) INV; try exact (STm.result_neq H _); simpl in H;
    pose proof (solution_monotone Hincr) as hm;
    hsplit;
    subst_simpl.
    destruct (ceq_or_cne _) eqn: CEQ; hsplit; subst_simpl;
    (destruct (two_const_ranges _ _) as [ ( (i, j) & CRi & CRj ) | ];
      [
        apply is_const_elim in CRi; destruct CRi as [ CRi CRi' ];
        apply is_const_elim in CRj; destruct CRj as [ CRj CRj' ];
        destruct (sumbool_of_bool _)
      |
      ]);
    hsplit; subst_simpl;
    flatten_eq hsplit; subst_simpl;
    (split; [
    | refine (fresh_lc_invariant _ _ _ _); try discriminate; auto;
      clear ω β hβ hv;
      intros ω β hβ
    ]);
    repeat match goal with
    | H : _ ! ?x = Some ?v |- _ =>
      let q := fresh "q" in
      let hq := fresh "hq" in
      apply get_env_Some_simpl in H;
      destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ H) as (q & hq & _);
      try (pose proof (hv x (lc_of_value v) (eq_trans (get_env_gso _ _ (lt_ne (get_env_inv _ _ H))) H) q hq)); clear H
  | H : lc_of_value ?v = (_ , _), K : context[ lc_of_value ?v ] |- _ =>
    rewrite H in K; simpl in K
  | CEQ: ceq_or_cne ?c = true, H: context[ cpair ?c ] |- _ =>
    destruct c; try (now clear -CEQ; vm_compute in CEQ); simpl in H; hsplit; subst_simpl
    end; subst; eauto;
    [
      repeat match goal with
      | H : project ?a ?b = Some ?c, K : ∀ _, _ → _ |- _ =>
        specialize (K c (project_in H)); subst
      end;
      pose proof (hv _ _ (get_env_gss _ _ _)) as Hw; simpl in Hw;
      specialize (Hw _ (lc_of_f_denote _ hβ)); rewrite <- Hw; clear Hw;
      rewrite Int.repr_unsigned;
      clear -CEQ;
      destruct c; vm_compute in CEQ; try easy; destruct s; reflexivity
    |
      simpl; erewrite lc_of_f_denote; eauto; easy
    | | | | | | ];

    generalize (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)); simpl;
    rewrite ! (LC.den_sub F); intros Hw';
    assert (∀ p, p < Pos.succ (Pos.succ p))%positive as X by (clear; intro; Psatz.lia);
    pose proof (π₃ hβ _ (X _)) as Y;
    simpl in Y;
    repeat (rewrite M.gso, M.gss in Y; [ | clear; Psatz.lia ]);
    specialize (Y _ _ _ eq_refl eq_refl eq_refl); clear X;
    setoid_rewrite (LC.den_sub F) in Y;
    (rewrite (LC.den_single F), (project_one' (proj1 hβ _ _ (proj2 (md_invariant _)))) in Hw'
    || rewrite (LC.den_single F), (project_one' (proj1 hβ _ _ (proj2 (md_invariant _)))) in Y);
    rewrite (LC.den_zero F) in Y;
    apply (field_integral F) in Y;

    try (pose proof (hv _ _ (get_env_gss _ _ _)) as Hw);

    destruct Y as [ Y | Y ] ;
    clean_field;
    try rewrite Y in Hw' ; try (setoid_rewrite Y in Hw; specialize (Hw _ project_zero); rewrite <- Hw).

      cut (Int.eq (Int.repr (ρ a)) (Int.repr (ρ b)) = false).
      clear; destruct s; intros X; simpl; rewrite X; exact eq_refl.
      refine (Int.eq_false _ _ _).
      refine (not_eq_sym _).
      refine (repr_inj _ _); eauto.
      refine (project_inj hq0 hq _).
      clear -Hw' F.
      intros EQ. rewrite EQ in Hw'. clear EQ.
      clean_field.

      clean_field.
      setoid_rewrite <- Hw' in Hw.
      specialize (Hw _ project_one). rewrite <- Hw.
      clear Hw Hw'.
      cut (Int.eq (Int.repr (ρ a)) (Int.repr (ρ b)) = true).
      clear; destruct s; intros X; simpl; rewrite X; exact eq_refl.
      setoid_rewrite Y in hq. setoid_rewrite (project_fun hq hq0 e). exact (Int.eq_true _).
      apply Pos.le_1_l.
      apply Pos.le_1_l.
      simpl in *.
      setoid_rewrite <- Y in Hw. specialize (Hw _ project_one). rewrite <- Hw. clear Hw.
      cut (Int.eq (Int.repr (ρ a)) (Int.repr (ρ b)) = false).
      clear; destruct s; intros X; simpl; rewrite X; exact eq_refl.
      refine (Int.eq_false _ _ _).
      refine (not_eq_sym _).
      refine (repr_inj _ _); eauto.
      refine (project_inj hq0 hq _).
      intros EQ. setoid_rewrite EQ in Hw'. clear EQ.
      clean_field.
      rewrite Hw' in Y. clean_field.
      clean_field. setoid_rewrite Hw' in Hw.
      specialize (Hw _ project_zero). rewrite <- Hw. clear Hw.
      cut (Int.eq (Int.repr (ρ a)) (Int.repr (ρ b)) = true).
      clear; destruct s; intros X; simpl; rewrite X; exact eq_refl.
      setoid_rewrite Y in hq. setoid_rewrite (project_fun hq hq0 e). exact (Int.eq_true _).
      apply Pos.le_1_l.
      apply Pos.le_1_l.

      rewrite Y, project_zero; intro; hsplit.
      clean_field. rewrite <- Hw', project_one; intro; hsplit.
      apply Pos.le_1_l.
      apply Pos.le_1_l.
      rewrite <- Y, project_one; intro; hsplit.
      clean_field. rewrite Hw', project_zero; intro; hsplit.
      apply Pos.le_1_l.
      apply Pos.le_1_l.
      apply INV; eauto.
      apply INV; eauto.
      apply Pos.le_1_l.
      apply Pos.le_1_l.
      apply INV; eauto.
      apply INV; eauto.
      apply Pos.le_1_l.
      apply Pos.le_1_l.
Qed.

Lemma init_sound :
   ∀ ω : LC.var, invariant ω (Arith.init (cQAP f_of_z)).
Proof.
    clear. unfold Arith.init.
    intros; simpl in *; repeat split; intros; try rewrite !get_env_empty in *; hsplit; simpl in *.
    - rewrite M.gempty in H0. discriminate.
    - rewrite M.gempty in H0. discriminate.
    - rewrite M.gempty in H0. discriminate.
    - rewrite M.gempty; reflexivity.
Qed.

Lemma input_sound :
   ∀ (x : input_data) (r : option (int * int)) (σ : cqap f)
   (y : positive) (σ' : cqap f) (Hincr : incr σ σ'),
   Arith.input (cQAP f_of_z) x r σ = STm.Result y σ' Hincr
   → (∀ ω : LC.var, invariant ω σ) → ∀ ω : LC.var, invariant ω σ'.
Proof.
    clear -F project_proper project_in project_inj project_zero project_one π₃ project_const project_one'.
    simpl. unfold fresh_lc, fresh_eqn, fresh_wire.
    intros x r σ y σ' Hincr H INV.
    pose proof (solution_monotone Hincr) as hm.
    flatten hsplit; subst.

    - refine (fresh_lc_invariant _ _ _ INV); try discriminate; intros ω β hβ; simpl; rewrite (lc_of_f_denote _ hβ); intro; hsplit.

    - simpl. intros ω β hβ w. simpl. split; [ |split]; auto.
      * intros v H. rewrite get_env_sspec in H. destruct (equiv_dec _ _) as [ -> | NE].
        + hsplit. simpl in H. hsplit. subst.
          match goal with |- context[ next_wire ?σ ] =>
            generalize ((proj1 hβ) (next_wire σ))
          end. simpl. rewrite M.gss. intros Q. specialize (Q _ eq_refl).
          destruct (not_none Q) as [p1 Q'].
          exists p1; simpl in *.
          unfold lc_of_wire. rewrite (LC.den_single F), right_identity;
            [ |destruct hβ as (_ & _ & H30); simpl in H30; clear -H30; Psatz.lia].
          split; auto.
          intros bs Hbs. rewrite M.gss in Hbs. discriminate.
        + generalize (proj1 (INV _ _ (hm _ _ hβ) w) v H). rewrite M.gso; auto.
      * intros bs Hbs.
        destruct (proj1 (proj2 (INV _ _ (hm _ _ hβ) w)) bs Hbs) as (H30 & Hval & val & p1 & p1' & H31 & H32 & H33 & H34).
        split; auto.
        split; auto.
        rewrite get_env_sspec. destruct (equiv_dec _ _) as [EQ|NE]; simpl.
        + rewrite EQ in H31. unfold get_env in H31.
          rewrite e0 in H31; try discriminate.
          clear. Psatz.lia.
        + exists val, p1, p1'.
          repeat split; auto.
      * intro H. apply (proj2 (proj2 (INV _ _ (hm _ _ hβ) w))).
        clear -H. Psatz.lia.

    - simpl in *.
      intros ω β hβ w. simpl.
      split; [ |split].
      * destruct (input_bit_spec H0 INV ω) as [H11 H12].
        intros v H. rewrite get_env_sspec in H. destruct (equiv_dec _ _) as [ -> | NE].
        + hsplit. simpl in H. hsplit. subst.
          simpl; auto.
          destruct (not_none (H11 _ hβ)) as [p1 Q'].
          rewrite M.gss. exists p1. split; auto. discriminate.
        + generalize (proj1 (H12 _ hβ _) _ H). rewrite M.gso; auto.
      * destruct (input_bit_spec H0 INV ω) as [H11 H12].
        intros bs Hbs.
        destruct (proj1 (proj2 (H12 _ hβ w)) bs Hbs) as (H30 & Hval & val & p1 & p1' & H31 & H32 & H33 & H34).
        split; auto.
        rewrite get_env_sspec. destruct (equiv_dec _ _) as [EQ|NE]; simpl.
        + rewrite EQ in H31. unfold get_env in H31.
          rewrite e1 in H31; try discriminate.
          clear. Psatz.lia.
        + split; auto.
          exists val, p1, p1'.
          repeat split; auto.
      * destruct (input_bit_spec H0 INV ω) as [H11 H12].
        intro H. apply (proj2 (proj2 (H12 _ hβ w))).
        clear -H. Psatz.lia.
Qed.


Lemma output_sound :
   ∀ (x : positive) (σ : cqap f) (y : positive) h (σ' : cqap f)
   (Hincr : incr σ σ'),
   Arith.output (cQAP f_of_z) x σ = STm.Result (y, h) σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.repr (ρ y) = Int.repr (ρ x) ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -π₃ project_proper F f_of_z_zero project_in project_const project_zero project_one project_add project_muln project_fun project_inj project_subrange. simpl.
    unfold QAP.get_range, fresh_lc, fresh_eqn, fresh_wire, QAP.get_lc.
    intros x σ y h σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    destruct (cqap_monotone_precise Hincr ρ hβ hv) as [hm1 hv1].
    hsplit.
    pose proof (solution_monotone Hincr''1) as hm''1.
    subst_simpl. destruct (sumbool_of_bool _); hsplit. subst_simpl.
    case_eq ((env s''0) ! x); [intro v| ]; intro Heq; rewrite Heq in H0; [ |hsplit].
    assert (H0': STm.Result (match v with | inl (_, r) => r | inr bs => range_of_binary_repr bs end) s''0 (incr_refl s''0) = STm.Result (E:=Errors.errmsg) x0 s'' Hincr''0) by (clear -H0; revert H0; case v; auto; intros [_ r]; auto).
    clear H0. hsplit. subst_simpl.
    case_eq (leb match v with | inl (_, r) => r | inr bs => range_of_binary_repr bs end (range_of_bit_numbers Int.wordsize)); intro Heq2; rewrite Heq2 in H5; simpl in *.
    - hsplit. subst_simpl.
      rewrite Heq in H0. hsplit. subst_simpl.
      split.
      * replace (ρ (next_var s''1)) with (ρ x); auto.
        refine (hv _ _ (get_env_gss _ _ _) _ _). simpl.
        generalize (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
        rewrite (LC.den_zero F), (LC.den_sub F).
        intro H. ring_simplify in H. symmetry in H.
        apply (field_zero_diff_eq F) in H. rewrite <- H. clear H.
        pose proof (get_env_Some_simpl _ _ Heq) as H'.
        rewrite H0 in H'.
        destruct (proj1 (INV _ _ hm1 _) _ H') as (q & hq & _).
        simpl in hq.
        generalize (hv1 x _ H' q hq).
        intros <-. auto.
      * clear hm1 hv1 ω β hβ hv.
        refine (fresh_lc_invariant _ _ _ _); try discriminate;
        intros ω β hβ; simpl;
        hsplit; subst_simpl.
        + generalize (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
          rewrite (LC.den_zero F), (LC.den_sub F).
          intro H. ring_simplify in H. symmetry in H.
          apply (field_zero_diff_eq F) in H. rewrite <- H. clear H.
          pose proof (get_env_Some_simpl _ _ Heq) as H'.
          rewrite H0 in H'.
          destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ H') as (q & hq & _).
          simpl in hq. rewrite hq. discriminate.
        + apply INV, (hm _ _ hβ).
    - hsplit. subst_simpl.
      split.
      * destruct (get_binary_decomposition_spec _ H0 INV ω) as [H100 H105].
        destruct (H105 β (hm''1 _ _ hβ)) as [H104 [H101 [[lc r] [Hincr'0 [H102 [p [p' [H17 [H18 H19]]]]]]]]]. clear H105.
        simpl in *. clear H0.
        unfold QAP.get_lc in H102. hsplit. subst_simpl.
        rewrite Heq in H102. hsplit. subst_simpl. clear EQ.
        destruct (@truncate_32_spec _ _ _ (hm''1 _ _ hβ) x0 (size_range_of_binary_repr_wordsize _ H104) H101) as (H110 & q & q' & H111 & H112 & H113).
        rewrite H18 in H111. inversion H111. subst p'. clear H111.
        generalize (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
        rewrite (LC.den_zero F), (LC.den_sub F).
        intro H. ring_simplify in H. symmetry in H.
        apply (field_zero_diff_eq F) in H.
        generalize (hv _ _ (get_env_gss _ _ _) q'). simpl.
        intro H1. rewrite <- H1; [ |rewrite <- H; auto].
        rewrite <- H113, <- H19.
        assert (H42: p = ρ x); [ |rewrite H42; reflexivity].
        pose proof (get_env_Some_simpl _ _ Heq) as H'.
        rewrite H102 in H'.
        generalize (hv1 x _ H' _ H17). auto.
      * clear hm1 hv1 ω β hβ hv.
        refine (fresh_lc_invariant _ _ _ _); try discriminate;
        intros ω β hβ; simpl;
        hsplit; subst_simpl;
        destruct (get_binary_decomposition_spec _ H0 INV ω) as [H100 H105];
        destruct (H105 β (hm''1 _ _ hβ)) as [H104 [H101 [[lc r] [Hincr'0 [H102 [p [p' [H17 [H18 H19]]]]]]]]]; simpl in *; clear H0 H105;
          unfold QAP.get_lc in H102; hsplit; subst_simpl;
          rewrite Heq in H102; hsplit; subst_simpl; clear EQ;
          destruct (@truncate_32_spec _ _ _ (hm''1 _ _ hβ) x0 (size_range_of_binary_repr_wordsize _ H104) H101) as (H110 & q & q' & H111 & H112 & H113);
          rewrite H18 in H111; inversion H111; subst p'; clear H111.
        + generalize (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)).
          rewrite (LC.den_zero F), (LC.den_sub F).
          intro H. ring_simplify in H. symmetry in H.
          apply (field_zero_diff_eq F) in H. rewrite <- H. clear H.
          rewrite H112. discriminate.
        + apply H100, (hm''1 _ _ hβ).
Qed.


Lemma assert_sound :
   ∀ (x: positive) y (σ σ' : cqap f) (Hincr : incr σ σ'),
   Arith.assert (cQAP f_of_z) x σ = STm.Result y σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.eq (Int.repr (ρ x)) Int.zero = false
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -F project_in project_inj project_const project_zero project_one project_one' project_proper π₃.
    simpl. unfold QAP.get_lc, fresh_eqn, fresh_wire.
    intros x y σ σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. subst.
    match type of H with context[ if eqb ?a ?b then _ else _] =>
      destruct (eqb a b) as [ Hbit | Hnbit ]; hsplit; subst;
      destruct (_ ! _) eqn: Henv; hsplit; subst_simpl
    end.

    - (* bit *)
    pose proof (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)) as K.
    unfold lc_one in *.
    rewrite (LC.den_zero F), (LC.den_sub F), (LC.den_single F) in K; try apply Pos.le_1_l. clean_field.
    rewrite (project_one' (proj1 hβ _ _ (proj2 (md_invariant _)))) in K.
    specialize (hv _ _ (get_env_Some_simpl _ _ Henv) (Zpos xH)); clear Henv.
    rewrite H0 in hv. simpl in hv. rewrite <- K in hv.
    specialize (hv (project_one)).
    split.
    refine (Int.eq_false _ _ _). rewrite <- hv. discriminate.
    clear ω β hβ K.
    intros ω β hβ. exact (INV ω β (hm _ _ hβ)).

    - (* maybe not a bit *)
    match type of H with context[ if ?b then ?t else ?e ] =>
      remember (if b then t else e) as x'r' eqn: Hx'r'
    end.
   destruct x'r' as [[x' r'] s].
   hsplit. subst_simpl.
    pose proof (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)) as K.
    unfold lc_of_wire, lc_one in K.
    assert (next_wire s''0 <= ω)%positive.
    generalize (proj2 (proj2 hβ)). simpl. clear. ppsimpl. Psatz.lia.
    destruct (equiv_dec (β xH) 1 ) as [ EQ | NEQ ].
    2:
    destruct (not_none (proj1 hβ  _ _ (proj2 (_.(md_invariant))))) as ( one & hone );
    elim (project_inj hone project_one NEQ);
    apply project_in in hone; exact (Interval.const_exact _ _ hone).

    rewrite ! (LC.den_single F) in K;
    auto using Pos.le_1_l.
    rewrite (project_one' (proj1 hβ _ _ (proj2 (md_invariant _)))) in K. clean_field.
    apply (field_prod_nz F) in K.
    destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ (get_env_Some_simpl _ _ Henv)) as (q & hq & _).
    specialize (hv _ _ (get_env_Some_simpl _ _ Henv) q hq); clear Henv. subst q.
    repeat match goal with
  | H : lc_of_value ?v = (_ , _), K : context[ lc_of_value ?v ] |- _ =>
    rewrite H in K; simpl in K
    end.
    pose proof (project_in hq) as hrange.
    split.
    flatten_eq hsplit; subst.
    rewrite (LC.den_sub F), (LC.den_single F) in K; auto using Pos.le_1_l.
    rewrite EQ in K. clean_field.
    apply (field_zero_diff_neq F) in K.
    pose proof (project_inj hq (project_const Int.modulus) K) as hnz.
    apply Int.eq_false, repr_inj'.
    clear hq K hm hβ.
    intros Hx. change (0 mod Int.modulus) with Z0 in Hx.
    generalize (Interval.leb_denote _ _ e _ (Interval.sub_correct _ _ _ _ hrange (Interval.const_correct _))).
    clear - hnz Hx.
    unfold Interval.denote, is_true. simpl. rewrite Bool.andb_true_iff, ! Z.leb_le.
    change Int.max_unsigned with 4294967295.
    change Int.modulus with 4294967296 in *.
    ppsimpl.
    Psatz.lia.
    pose proof (project_inj hq project_zero K) as hnz.
    apply Int.eq_false, repr_inj.
    generalize (Interval.leb_denote _ _ e _ hrange). clear.
    unfold Interval.denote, is_true. simpl. rewrite Bool.andb_true_iff, ! Z.leb_le.
    change Int.max_unsigned with 4294967295. Psatz.lia.
    exact hnz.
    clear - INV hm.
    intros ω β hβ. exact (INV ω β (hm _ _ hβ)).
Qed.

Lemma assert_eq_sound :
   ∀ (x y : positive) (σ σ' : cqap f) (Hincr : incr σ σ'),
   Arith.assert_eq (cQAP f_of_z) x y σ = STm.Result tt σ' Hincr
   → ∀ ρ : positive → Z,
     cqap_solution σ' ρ
     → (∀ ω : LC.var, invariant ω σ)
       → Int.eq (Int.repr (ρ x)) (Int.repr (ρ y)) = true
         ∧ (∀ ω : LC.var, invariant ω σ').
Proof.
    clear -F project_one' project_fun project_proper π₃.
    simpl. unfold QAP.get_lc, fresh_eqn.
    intros x y σ σ' Hincr H ρ (β & ω & hβ & hv) INV.
    pose proof (solution_monotone Hincr) as hm.
    hsplit. subst. flatten_eq hsplit. subst_simpl.

    pose proof (π₃ hβ _ (Pos.lt_succ_diag_r _) _ _ _ (M.gss _ _ _) (M.gss _ _ _) (M.gss _ _ _)) as K.
    unfold lc_one in K.
    rewrite (LC.den_single F) in K; try apply Pos.le_1_l.
    rewrite (project_one' (proj1 hβ _ _ (proj2 (md_invariant _)))) in K. clean_field.
    destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ (get_env_Some_simpl _ _ Heqo)) as (q & hq & _).
    destruct (proj1 (INV _ _ (hm _ _ hβ) _) _ (get_env_Some_simpl _ _ Heqo0)) as (q' & hq' & _).
    rewrite <- (hv _ _ (get_env_Some_simpl _ _ Heqo) _ hq).
    rewrite <- (hv _ _ (get_env_Some_simpl _ _ Heqo0) _ hq').
    repeat match goal with
  | H : lc_of_value ?v = (_ , _), K : context[ lc_of_value ?v ] |- _ =>
    rewrite H in K; simpl in K
    end.
    rewrite K in hq.
    pose proof (project_fun hq hq' e). subst q'.
    refine (conj (Int.eq_true _) _).
    clear ω β hβ K hv hq hq'.
    intros ω β hβ. exact (INV ω β (hm _ _ hβ)).
Qed.

Theorem cQAP_sound : Arith.spec (cQAP f_of_z) cqap_solution (λ q, ∀ ω, invariant ω q).
Proof.
split.
- exact of_int_sound.
- exact to_int_sound.
- exact add_sound.
- exact sub_sound.
- exact mul_sound.
- exact land_sound.
- exact lor_sound.
- exact lxor_sound.
- exact shru_sound.
- exact cmp_sound.
- exact init_sound.
- exact input_sound.
- exact output_sound.
- exact assert_sound.
- exact assert_eq_sound.
Qed.

End FIELD.

Arguments cqap_solution {f feq _ _ _ _ _ _} _ _ _.
Arguments cqap_monotone {_ _ _ _ _ _ _ _} _ [_ _] _ _ _.
